/*****************************************************
  fichier : c_file.cpp
  usage   : classe extension de FILE*
  date    : 1991--
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  : ideas from wxWindows (ungetch)
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/


#include <c_defs.h>
#include <c_file.h>



/* pour c_autodependency */
#include    <fcntl.h>
#include    <sys/types.h>
#include    <sys/stat.h>	    /* access*/
#ifndef   M_UNIX
#include    <io.h>
#else
#define		O_BINARY	0
#endif


/*------------------------------------*/
//  getch() avec CRLF2LF conversion
/*------------------------------------*/


bool c_File::open(lpcstr filename, char mode) 
/*  Toujours en mode BINAIRE pour grer CRLF2LF
/*  Retour:
/*      vrai fichier ouvert ou existe
/*      faux: fichier n'existe pas
*/
{
char  realmode[3];  /* Car le paramtre est toujours un caractre */
char	flag=0; 	    /* >0 si checking-mode */

    if( !c_isvalid(filename)) return false; 
    m_name= filename;

    if( 0!= (flag=( 'x'==mode)?true:false) )    /* compatibilite */
    {
      // TODO: checking du file
	    mode= 'r';
    }

    realmode[0]= mode;
    realmode[1]= 'b';   // toujours BINAIRE
    realmode[2]= 0;
    m_fp=fopen(c_fcheck(m_name), realmode);

    return m_fp ? true : false;     // throw ?
}


size_t c_File::length() const
{
    c_File& self = *(c_File*)this;   // const_cast

    size_t posOld = ftell(self.fp());
    if ( posOld != (size_t)-1 )
    {
        if ( self.seekend() )
        {
            size_t len = getpos();

            (void)self.seek(posOld);

            return len;
        }
    }

    return (size_t)-1;
}


/*------------------------------------------------------------*/
// Lecture avec gestion de buffer dynamique pour ungetch
/*------------------------------------------------------------*/

uchar c_File::getch()
{ 
  register int ch;

    if( -1!= (ch= _getback()))
      return ch;

#ifdef _WIN32
    if( CR==(ch=fgetc(m_fp)))
  	  ch=fgetc(m_fp);		    /* CRLF2LF conversion   */
#else
  	  ch=fgetc(m_fp);		    /* UNIX standard */
#endif

    return(ch);
}

int c_File::read(lpstr buffer, int size)  
{
    size_t retsize= 0;
    uchar ch;
    while( (size>0) && !eof())
    {
      *buffer++= ch= getch();
#ifdef _WIN32
      retsize+= (NL==ch)?2:1;   // Corriger NL!!
      size   -= (NL==ch)?2:1;   // Corriger NL!!
#else
      retsize++;
      size   --;
#endif

    }
    return retsize;             // total lu
}



int c_File::_getback()
// Retourne -1 si la queue est vide
{
    if (!m_wback) return -1;               // vide 
    int retch= m_wback[m_wbackcur++];
    if( m_wbackcur== m_wbacksize)
    {
        free(m_wback);
        m_wback = (char *)NULL;
        m_wbacksize = 0;
        m_wbackcur = 0;
    }
    return retch;
}

lpstr c_File::_allocback(size_t count)
// Allouer de l'espace ungetch()
{
    /* bytes en attente wback buffer */
    size_t toget = (m_wback) ? m_wbacksize - m_wbackcur : 0 ;

    /* allocate a buffer large enough to hold prev + new data */
    lpstr tempbuf= (char *) malloc(count + toget);
    if (!tempbuf) return NULL;

    /* copy previous data (and free old buffer) if needed */
    if (m_wback)
    {
        memmove(tempbuf + count, m_wback + m_wbackcur, toget);
        free(m_wback);
    }

    /* done */
    m_wback = tempbuf;
    m_wbackcur  = 0;
    m_wbacksize = count + toget;

    return m_wback;
}

/*------------------------------------*/
/*  Fonctions de deplacement
/*------------------------------------*/

bool c_File::seek_not_of( lpcstr szChars)
/*	true =	1er char non exclus (ungetch)
/*  false=	eof()
*/
  {
    while ( !eof())
    {
      if( !strchr(szChars, peek()) )
        return true;
      else
        getch();
    }
    return false;
  }

bool c_File::seek_no_blanks()
/*	true =	1er char > SPC
/*  false=	eof()
*/
  {
    while ( !eof())
    {
      if( SPC < peek() )    // 
        return true;        // pointe caractre > SPC (espace NON compris)

      getch();         
    }
    return false;
  }

bool c_File::seekeol()
/*	true =	APRES NL == debut de ligne
/*  false=	eof()
*/
  {
      for( uchar ch= getch(); (NL!=ch)  && !eof(); ch= getch() )
        ;
      return (NL==ch);	    /* apres NL */
  }

bool c_File::seekline( int lines)
/*  Positionne le fichier avec m_FileNo==line
/*  C_REWIND SYSTEMATIQUE (prevoir +sieurs files ouverts (=edb!))
/*  RETOUR:
/*	true =	ok
/*  false=	eof()
*/
{
    rewind();
    while( !eof() && lines--)
	    if( !seekeol()) return false;		/* eof unexpected! */

    return lines ? false : true;
}


bool c_File::seektag( lpcstr userTag, c_Str* echo)
/*  Action:
/*	Recherche la premiere ligne commencant par tag.
/*  (pour une recherche dans le texte-> SeekStr())
/*  
/*  Les espaces devant tag sont ignoreds.
/*  Le texte du Tag est insensible  la casse
/*	(L'espace devant <szTag> ne fait pas partie de la comparaison)
/*
/*  Retour:
/*	vrai  = pointe APRES le szTag (sur la meme ligne)
/*	faux  = eof()
/*
/*  Exemples:
/*	    seektag("url:")
/*
*/
{
register ch; 
lpcstr szTag= userTag;
lpcstr saveTag= szTag;

    if( eof())   return false;

    // Analyse du Tag a rechercher
    if( !c_isvalid(szTag)) 
        return false;           /* aucun sens	*/

    szTag= c_adr_no_blank(szTag);    /* pas de blanc_!	*/

    if( !c_isvalid(szTag)) 
        return false;           /* aucun sens	*/

    // Lecture du fichier
    while( !eof())
    {
	    while( NL!= (ch=getch()))     // Chercher dbut de ligne
        if( echo)  echo->append(1,ch);   // Output 

      if( eof()) break;
      if( echo)  echo->append(1,ch);     // Output NL


	    while( SPC >= (ch=getch()))   // Chercher premier char
        if( echo)  echo->append(1,ch);        // Output 

      if( eof()) break;

      // Comparaison du TAG NON ECHOED!!
      ungetch(ch);                 // resync premier char != SPC
      saveTag= szTag;              // saveTag
      while ( *szTag)
      {
        register c1= toupper(*szTag);
        register c2= getch();
        if ( c1 != toupper(c2) )
          break; // no match
        
        szTag++;  
      }

      if( !*szTag)              // Tout le tag est lu ?
        return true;            // le TAG lui-mme n'est pas echoed

      if( echo) 
      {
        echo->append(saveTag,0,szTag-saveTag);   /* Output de ce qui est lu 	*/
        echo->append(1,ch);                      /* Output du dernier char  	*/
      }

      szTag = saveTag;          // restore et recommence

    }

    return false;               // eof()
}



/* version c_Str&*/
/*  Usage:
/*		Se dplace  la suite de la premire occurence de Sequence
/*		Commence  la position du fichier
/*
/*		Si Echo!=NULL, 2 recopie les caracteres de this
/*    SAUF la squence elle-meme (dja connue)
/*
/*  Retour:
/*		vrai: c_File pointe premier char APRES sequence 
/*          _Echo  ne contient PAS la sequence finale
/*		faux eof,	ou Sequence invalide
/*
/*  Exemples:
/*	SeekStr ( "#{\footnote ")
*/
bool c_File::_seekstr( lpcstr Sequence, c_Str* Echo)
{
lpcstr 	szP;
char	  ch=0;

    if( eof())			return false;
    if( !c_isvalid(Sequence))	return false;     /* aucun sens	*/

    while( !eof() )
    {
	  /* Rechercher le premier caractere valable */
	    szP= Sequence;
	    for( ch=getch(); *szP != ch; ch= getch() )
	    {
	        if(eof())  return false;	    /* eof!		*/
	        if( Echo) Echo->append(1,ch);        /* Output 	*/
	    }

    /* 
	  /* Sauver la position APRES le premier match 
	  /* comparaison avec Sequence 
	  */
	    do
	    {
          szP++;
	        ch = getch();
	        if(eof()) break;	/* eof! le match peut bien se trouver a eof */
	    } 
      while( *szP == ch);		/* Retester le premier match */

	    if(!*szP)				      /* tout le tag compared?	*/
      {
	        ungetch( ch);			/* unget le mismatch		*/
    	    return true;			/* Pointe aprs Sequence, non recopiee) */
      }

	  /* Sinon: ungetch le mismatch, 
    /* puis la sequence SAUF 1er char ! (qui est dj test)
	  */
	    ungetch( ch);
	    ungetch( Sequence+1, szP- (Sequence+1) );
	    if( Echo) Echo->append(1, *Sequence); /* Output 	*/
    }

/* PAS trouved: UNDEF */
    return false;	              /* pas trouved */
}

/* version c_File&, C'est exactement la mme syntaxe :(( */
bool c_File::seekstr( lpcstr Sequence, c_File& Echo)
{
lpcstr 	szP;
char	  ch;

    if( eof())			return false;
    if( !c_isvalid(Sequence))	return false;     /* aucun sens	*/

    while( !eof() )
    {
	  /* Rechercher le premier caractere valable */
	    szP= Sequence;
	    for( ch=getch(); *szP != ch; ch= getch() )
	    {
	        if(eof())  return false;	    /* eof!		*/
	        Echo<< ch;        /* Output 	*/
	    }

    /* 
	  /* Sauver la position APRES le premier match 
	  /* comparaison avec Sequence 
	  */
	    do
	    {
          szP++;
	        ch = getch();
	        if(eof()) break;	/* eof! le match peut bien se trouver a eof */
	    } 
      while( *szP == ch);		/* Retester le premier match */

	    if(!*szP)				      /* tout le tag compared?	*/
      {
	        ungetch( ch);			/* unget le mismatch		*/
    	    return true;			/* Pointe aprs Sequence, non recopiee) */
      }

	  /* Sinon: ungetch le mismatch, 
    /* puis la sequence SAUF 1er char ! (qui est dj test)
	  */
	    ungetch( ch);
	    ungetch( Sequence+1, szP- (Sequence+1) );
	    Echo<< *Sequence; /* Output 	*/
    }

/* PAS trouved: UNDEF */
    return false;	              /* pas trouved */
}


/*------------------------------------*/
/*  Fonctions de Lecture 
/*------------------------------------*/

uchar c_File::peek()
{
  uchar ch= getch();
  ungetch(ch);
  return ch; 
}

int c_File::peek(lpstr buf, size_t count)
{
    int nread= read   (buf, count);
    ungetch(buf, count);
    return nread; 
}

// Lecture jusqu' EOL non compris. 
// Retourne nombre de ch lus
int c_File::peekline(c_Str& buf)
{
c_Str temp;
    if( false==seekeol(temp)) return false;
    buf  << temp; 
    temp << NL;         // ungetch inclus NL !
    ungetch(temp);
    return temp.size();   
}

int c_File::ungetch(lpcstr buf, size_t bufsize)
  {
      lpstr pback= _allocback(bufsize);
      if (! pback) return 0;

      memcpy( pback, buf, bufsize);
      return bufsize;
  }

bool c_File::ungetch( char ch)
  {
      lpstr pback= _allocback(1);
      if (! pback) return false;

      *pback= ch;
      return true;
  }

bool    c_File::getfile   (c_Str& buf)	  
{ 
  int count= buf.length();
  bool nret;
    buf.reserve(count);
    nret= (count== read((lpstr) buf, count)) ? true : false; 
    buf.sync();      // resynchronise m_end()
    return nret;
}
                            

long  c_File::getnumber(int count)     // longueur dtermine
  {
  long prmt;
  bool ishexa = ('0' == peek())? true : false;
    if( ! count)  
    {
      fscanf( fp(), (ishexa)?"%X":"%d", &prmt);
    }
    else
    {
    char buf[32]; 
	    read(buf, count);
      sscanf( buf, (ishexa)?"%X":"%d", &prmt);
    }
	  return prmt;
  }	

#if 0

/*- ------------------------------------------- ReadRTF
/*  Recherche le premier caractre n'appartenant PAS
/*	a un TAG de formatage RTF. 
/*  Les successions de Tags RTF separs par des espaces sont 
/*	correctement reconnues
/*  Le dernier caractre de la squence est
/*	- soit un espace de remplissage
/*	- soit un NUL de remplacement de NL
/*  A l'appel, le PREMIER caratre (sauf espaces) DOIT ETRE \ BSLASH
/*
/*  Entre:
/*	EStream szbuffer
/*  Sortie:
/*	this	    pointe sur le premier caractre significatif
/*	szbuf	    contient la squence de formatage RTF
/*	int	    donne le nombre de caractres lus
/*		    -1 si error
/*
/*
*/

int c_File::GetFormatsRTF(c_Str& szBuffer)
{
int	Code;
int	nCount=0;
    // Lecture ESPACES+RTFTags
	Code= getch();
	do
	{
	// Lecture ESPACES
    for( ; isspace(Code); nCount++, Code= getch())
		  if( !iscntrl(Code))	// ignorer contrles
		    szBuffer<< SPC;		// uniquement SPC
		
	// Vrifie si le premier caractere== BSLASH
	  if( BSLASH!=Code) 
		  break;

	// Lecture de token RTF {isalnum, MINUS, BSLASH}
    	// Admettre la syntaxe \* comme premire ocurrence
	  szBuffer<< Code;
	  Code= getch();
	  nCount++;
	  if( ASTERISK==Code || c_isrtf(Code) )
		do  // Lire le reste des tags RTF de manire standard
		{
		    szBuffer<< Code;
		    Code= getch();
		    nCount++;
		}
		while( c_isrtf(Code));
	}
	while( isspace(Code) );

    // Code delimiteur== !(espace ||isrtf()) non compris
    // Prcd de BSLASH?: ne pas inclure ce BSLASH (n'appartient pas a TAGs)
    // Ex.-	BSLASH ou LBRACE, cad un caractere literal?
    // Ex.-	TILDE, cad un espace non separateur?
    // Ex.-	QUOTE, cad un accent?
    // Ex.-	STAR , cad une extension Word?
    // Ex.-	QUESTION, cad un breakpoint? (propre au logiciel!)
	if( BSLASH == szBuffer.Last())	// BSLASH ==code prcdent?
	{
	    szBuffer.Truncate(szBuffer.Len()-1);		// Dpiler le dernier caractre!
	    SeekCur(-2);		// Synchronise sur BSLASH	    
	    nCount-=2;			// Pointe sur premier BSLASH
	}
	else
	{
	    ungetch(Code);		// Pointe 1er code significatif
	    nCount--;
	}
	return nCount;			// Nombre de caractres lus
}



int c_File::GetParagraphe( lpstr szBuffer, int nSize)
/*************************************/
/*  Usage:
/*	Lecture d'un bloc de texte
/*	Admet les comandes #line off|on
/*  #/PRE (default)
/*	Un paragraphe est prcd par 0 ou plus lignes vides
/*	Un paragraphe est suivit par 1 ligne vide ou eof
/*	Les tabs, cr et espaces dupliques sont ignores
/*  #PRE 
/*	La lecture se realise ligne a ligne
/*  #P	
/*	Interruption forcee de lecture
/*
/*  Entree:
/*	$1	szBuffer de reception
/*	$2	taille de szBuffer 
/*  Retour:
/*	0>=int	    Nbre de caracteres lus (=0 si eof)
/*		    NEWLINE n'est PAS retourned
/*		    NUL est appended, 
/*	int=nSize   NUL est appended, mais indique possible troncature
/*	
/*	-1	Erreur
/*  Remarques:
/*	Implemente pour parsing de fichiers texte
/*	Les formatages avec TAB/SPC sont ignoreds (ramenes a un seul espace)
*/
{
int Code;
int OldCode;
int nCount; 
boole bLineMode= off;

    nCount= 0;
    szBuffer[0]=0;

/* Ignorer lignes vides prcdentes
*/
    do
	Code=getch();
    while( NEWLINE==Code && eof!=Code);

    if( eof== Code) 
	return nCount;
    else
	Ungetc(Code);

/* Boucle de lecture de paragraphe
*/
    OldCode=NEWLINE;
    while( (nCount<nSize) && (eof!=(Code=getch())))
    {

    // Intrprete de commandes
	if( DIESE==Code && NEWLINE==OldCode)	// tester NEWLINE+DIESE
	{
        char    szMeta[32];
		GetLine(szMeta);
		c_strupper(szMeta);
//		if(strequ( "/PRE", szMeta))	bLineMode=off;
		if(strequ( "PRE", szMeta))	bLineMode=on;
//		if(strequ( "P", szMeta))	{ LastPos(); break; }
		continue;
	}

	// Mode de lecture ligne a ligne?
	if( (NEWLINE==Code) && (on==bLineMode))  break; 

	// Mode de lecture de paragraphes
	if( (NEWLINE==Code) && (NEWLINE==OldCode))  break;

	if( isspace(Code) && isspace(OldCode) ) continue;

	OldCode= Code;			    // Conserver 

	// Conversion a SPC
        szBuffer[nCount]=(isspace(Code))? SPC:Code;
	nCount++;
    }

// Unget NewLine, pour resync
    if( eof!= Code) Ungetc( NEWLINE);
    if( nCount==nSize) nCount--;    // Ecraser le dernier char!
    szBuffer[nCount]=0;
    return nCount;
}
/****************************************/
#endif