/*****************************************************
  file    : c_Vector.cpp
  usage   : vecteurs de bits, chaines et objets
  date    : 2001
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  :
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/

#include <c_vector.h>


/*----------------------------------------------------*/
/*  c_Vector
/*----------------------------------------------------*/

#define C_VBLOCK 8    // Nombre de blocs allous d'un coup
#define C_NEWSIZE(b)  ( ((1 + ((b)/C_VBLOCK) ) * C_VBLOCK))

bool c_Vector::reserve(int units)
// retourne la longueur alloue
{
  if( 0== m_size)					// 1er setup?
  {						   
	  m_size= C_NEWSIZE(units);
	  m_data= (ulong*) malloc( m_size* sizeof(ulong));
	  m_count= 0;
    return true;
  }

  int  freesp= m_size- m_count;   // freespace? 
  if( ! m_size- m_count) 
  {
	  m_size+= C_NEWSIZE(units);
	  m_data= (ulong*) realloc( m_data, m_size* sizeof(ulong));
  }

  return true;
}

void  c_Vector::add     ( ulong item)
{
  reserve(1);
  m_data[ m_count++] = item; 
}

void c_Vector::insert  ( int pos, ulong item)
{
  if( pos < 0) pos+= m_count;
  if( pos > m_count) 
  {
    add(item);
    return;
  }

  reserve(1);
  memmove(m_data+pos, m_data+pos+1, (m_count-pos)*sizeof(ulong));
  m_data[ pos ] = item; 
}

int c_Vector::find( ulong val) const
{
  for( int loop=0; loop< m_count; loop++)
    if( val== m_data[ loop ])
      return loop; 
  return -1;
}



ulong c_Vector::item    ( int pos) const
{
  if( !m_count) return 0L;
  if( pos < 0) pos+= m_count;
  if( pos > m_count) return 0L;

  return m_data[ pos ];
}

void c_Vector::erase ( int pos, int len)
{
  if( !m_count) return;
  if( pos < 0) pos+= m_count;
  if( pos > m_count) return;
  if( !len) len= m_count;
  if( pos+len > m_count) 
  {
    m_count= pos;
    return;
  }
  memmove(m_data+pos+len, m_data+pos, (m_count-(pos+len))*sizeof(ulong));
}


void c_Vector::pack    ()
{
  if( !m_size) return;
  m_data= (ulong*) realloc( m_data, m_count*sizeof(ulong));
  m_size= m_count;
}


#undef C_VBLOCK 
#undef C_NEWSIZE



/*----------------------------------------------------------
/*  c_Mappe
/*----------------------------------------------------------*/

#define C_BYTOFFSET(pos)  (m_ubits * ( pos / 32))
#define C_SHIFTMASK(pos)  (m_ubits * ( pos % 32))


void c_Mappe::define( int unit=0) 
{ 
  if(!unit) unit= 1;
  m_ubits= unit;
  m_mask = (~0L) >> (32-m_ubits);   // complment de bit :))
}


bool c_Mappe::setmap(int pos, int val)
// positionne l'unit
{
  if( pos > m_elem) return false;
  
  int byte  = C_BYTOFFSET(pos);   
  ulong* p1 = m_data;
  
  ulong mask  = (val & m_mask) << C_SHIFTMASK(pos);
  p1[ byte] |= mask;   // Masque
  p1[ byte] &= mask;   // affecte

  return true;
}

bool c_Mappe::resmap(int pos)
// annule l'unit
{
  if( pos > m_elem) return false;

  ulong* p1 = m_data;
  int byte  = C_BYTOFFSET(pos);   

  ulong mask = (m_mask) << C_SHIFTMASK(pos);
  p1[ byte] |= mask;   // Masque
  p1[ byte] ^= mask;   // XOR 

  return true;
}

int c_Mappe::getmap(int pos) const
// valeur de l'unit
//  -1 si outbound
{
  if( pos > m_elem) return false;
    
  ulong* p1 = m_data;
  int byte  = C_BYTOFFSET(pos);   

  ulong mask = (m_mask) << C_SHIFTMASK(pos);
  return (p1[ byte] & mask) >> C_SHIFTMASK(pos);
}

int  c_Mappe::count ()
// nombre d'units NON NULLES!!!!
{
  int count=0;
  for( int loop= 0; loop< m_elem; loop++)
  {
    count+= (getmap(count) ? 1:0); 
  }
  return count;
}


c_Mappe& c_Mappe::operator|=(const c_Mappe& v)
{
  ulong* p1 = m_data;
  int count= c_min(m_elem, v.m_elem); 

  int pos= 0; 
  int val;

  while( count--)
  {
    val= v.getmap(pos);
    int byte   = C_BYTOFFSET(pos);   
    ulong mask = (val & m_mask) << C_SHIFTMASK(pos);
    p1[ byte] |= mask;   // Masque
  }

  return *this;
}

c_Mappe& c_Mappe::operator&=(const c_Mappe& v)
{
  ulong* p1 = m_data;
  int count= c_min(m_elem, v.m_elem); 

  int pos= 0; 
  int val;

  while( count--)
  {
    val= v.getmap(pos);
    int byte  = C_BYTOFFSET(pos);   
    ulong mask = (val & m_mask) << C_SHIFTMASK(pos);
    p1[ byte] &= mask;   // Masque
  }

  return *this;
}


c_Mappe& c_Mappe::operator^=(const c_Mappe& v)
{
  ulong* p1 = m_data;
  int count= c_min(m_elem, v.m_elem); 

  int pos= 0; 
  int val;

  while( count--)
  {
    val= v.getmap(pos);
    int byte  = C_BYTOFFSET(pos);   
    ulong mask = (val & m_mask) << C_SHIFTMASK(pos);
    p1[ byte] ^= mask;   // Masque
  }

  return *this;
}

/*  Ajout de bits
*/
c_Mappe& c_Mappe::operator<<(const int val)
{

  if( 0== C_SHIFTMASK(m_elem) )     // le prochain caractre est hors limites ?
    reserve( 1);                    // ajouter un caractre (64 en fait)

  setmap(m_elem, val);
  return *this;
}



/*----------------------------------------------------------*/

