/*****************************************************
  file    : c_filefn.cpp
  usage   : ANSI C extensions pour Windows principalement
  date    : 1991--
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  :
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/

/*------------------------------------*/
/*	Manipulation de noms de fichiers disques
/*------------------------------------*/


#include <c_defs.h>
#include <c_file.h>

/* pour c_autodependency */
#include    <fcntl.h>
#include    <sys/types.h>
#include    <sys/stat.h>	    /* access*/
#ifndef   M_UNIX
#include    <io.h>
#else
#define		O_BINARY	0
#endif


//#define "wx/utils.h"

/*------------------------------------*/
/*  Syntaxe de path
/*------------------------------------*/

C_DLLEXPORT lpstr c_fcheck( c_Str& p) { c_fcheck( (lpstr)p); p.sync(); return p; }
C_DLLEXPORT lpstr c_fcheck( lpstr p)
/*  RETOUR: (adresse) la meme avec correction de syntaxe DOS ou UNIX
*/
{
register lpstr  s=p;

#ifdef M_UNIX
    while( NIL != (s=strchr(s,BSLASH)) ) *s= FSLASH;
#else
    while( NIL != (s=strchr(s,FSLASH)) ) *s= BSLASH;
#endif

    s= c_adr_nul(p)-1;
    if( NL==*s ) *s--=NUL;  /* elimine /nl	*/

    return( p); 	          /* retourne la meme */
}


/*------------------------------------*/
/*  splitpath subroutines
/*------------------------------------*/


static lpstr _extname( lpstr  p)
/*  RETOUR: (adresse) de l'extension (si existe)
/*	    N'utilise PAS de buffer static
*/
{
register lpstr 	    s;
    c_fcheck( p);
    if( NIL != (s=strrchr( p,PERIOD)) )
	return(s+1);	    /* apres hash!  */
    else
	return("");	    /* chaine vide  */
}

static lpstr _filename( lpstr  p)
/*  RETOUR: (adresse) de basename, extension inclue (si existe)
/*	    N'utilise PAS de buffer static
*/
{
register lpstr 	    s;
    c_fcheck( p);
#ifdef M_UNIX
    if( NIL != (s=strrchr( p,FSLASH)) )
#else
    if( NIL != (s=strchr ( p,COLON )) ) p=s+1;	/* drive declaration? */
    if( NIL != (s=strrchr( p,BSLASH)) )
#endif
	    return(s+1);	    /* apres hash!     */
    else
	    return(p);	      /* le nom lui-meme */
}

static lpstr _basename( lpstr  p)
/*  RETOUR: (static) basename = filename SANS extension
/*	    static buffer
/*	    Une extension est defined par PERIOD
*/
{
static char basename[128];   // Suffisant ? 

register lpstr 	    s;
    memset ( basename,NUL,sizeof(basename));
    strncpy( basename, _filename(p),sizeof(basename)-1);
    if(NIL!=(s= strrchr( basename,PERIOD)))    /* tronque extension */
	  *s=NUL;
    return(basename);
}

static lpstr _pathtrunc( lpcstr  p)
/*  RETOUR: (static) truncname= path complet SAUF extension
/*	    static buffer
/*	    Une extension est defined par PERIOD
*/
{
static PATH_t pathname;
register lpstr 	s;

    strcpy  ( pathname,p);
    c_fcheck( pathname);
    if(NIL!=(s= strrchr( pathname,PERIOD)))    /* tronque extension */
	    *s=NUL;
    return(pathname);
}

static lpstr  _pathname( lpcstr  p)
/*  RETOUR: (static) pathname determined comme ce qui precede basename
/*	    - jamais NUL: au moins "."
/*	    - SANS SLASH terminal
/*
*/
{
static PATH_t pathname;
register lpstr 	s;

    strcpy  ( pathname,p);
    c_fcheck( pathname);
    s= _filename(pathname);	    /* adresse sur buffer!  */
    s[0]=NUL;			    /* tronquer sur SLASH ou apres COLON */

/* avec defined path */
    if( s>pathname)
	  if( BSLASH==s[-1] || FSLASH==s[-1])	/* apres SLASH?  */
	  {
	    s[-1]=NUL;		    /* eliminer xSLASH precedent */
	    return(pathname);	    /* si s==pathname: donne root path */
	  }

/* sans defined path? */
    s=(COLON==pathname[1])? pathname+2 : pathname;
    if( NUL==*s)	     /* NULpath?: au moins PERIOD */
    {
	    s[0]=PERIOD;
	    s[1]=NUL;
    }
    return( pathname);
}



C_DLLEXPORT lpstr  c_splitpath ( lpstr p, int options)
//#define	C_DIR	4	
//#define	C_NAME	2
//#define	C_EXT	1
{
    switch(options)
    {
    case C_EXT:	            return _extname(p);
    case C_DIR:	            return _pathname(p);
    case C_DIR + C_NAME:    return _pathtrunc(p);
    case C_NAME+ C_EXT:     return _filename(p); 
    case C_NAME:	          return _basename(p);
    }
    return NULL;
}

C_DLLEXPORT lpstr  c_makepath( lpcstr p, lpcstr b, lpcstr e)
/*  RETOUR: (static) pathname:
/*	parametres DOIVENT etre syntaxiquement correct
/*	Les FLSASH/PERIOD sont supplied
/* (p,b,e) parametres non altered (static buffer)
*/
{
static PATH_t _bufpath;

    if (NIL==p) p= "";	 /* sans SLASH terminal */
    if (NIL==b) b= "";
    if (NIL==e) e= "";
    if( e[0]=='.') e++;

    sprintf (_bufpath,"%s/%s.%s", p,b,e);
    c_fcheck(_bufpath);
    return  (_bufpath);
}

/*------------------------------------*/
/*  test subroutines
/*------------------------------------*/

C_DLLEXPORT bool c_fdependency(lpstr f1,lpstr f2)
/*  Compare dates f1=but et f2=source
/*  Penser en termes de makefiles:
/*	.f1 dependant de f2	## exemple .e.o:
/*	action si f2>f1 	## emake $*.e
/*  TRUE  si:	si f2 est POSTERIEUR a F1 (modification->dependance)
/*  FALSE si:
/*	- f2 ou f1 absent (ENOENT): retourne FALSE
/*	- f1 est posterieur a f2    (pas de dependance)
/*
/* NOTE
/*	Sous UNIX, O_BINARY est defined 0
/*	stat() !ANSI mais compatible UNIX
*/
{
struct	stat	statf1;
struct	stat	statf2;
int   retvalue;

/** test de dates */
    stat( f1,&statf1);
    stat( f2,&statf2);
	  retvalue= ((ulong)statf1.st_mtime > (ulong)statf2.st_mtime)? 0:-1;
    return retvalue ? true : false;
}

C_DLLEXPORT  bool c_fexists(lpstr  path)
{
    if( NIL== c_fcheck(path)) return(-1);
    return access( path,0) ? false : true;    // teste existence only (=0)
}

C_DLLEXPORT size_t c_fsize(lpstr  path)
{
static struct  stat	  filestat;
    if( stat( path, &filestat)) return(-1); /* erreur */
    return( filestat.st_size);
}

C_DLLEXPORT bool c_fcopy( lpstr inp, lpstr out, bool modeappend)
/*  Mode: 0=creat, !0=append (verifie si existe)
*/
{
int	finp,fout;	/* handlers */
int	size;
lpstr 	buffer;

/* tester size source */
    if( false == c_fexists(inp)) return false;
    if( -1== (finp= open(inp,O_BINARY|O_RDONLY)))  return false;


/* Append?: tester si out existe */
    c_fcheck(out);

    if( modeappend)
	    fout= open(out,O_APPEND|O_BINARY);
    else
	    fout= open(out,O_CREAT|O_TRUNC|O_BINARY|O_RDWR,S_IWRITE);

    if( -1==fout)
    {
	    close( finp);
	    return false;
    }

/* boucle de registres */
    buffer  = (lpstr )malloc(512);

    while( 0< (size= read(finp,buffer,512)) )
    	write(fout,buffer,size);

    close(fout);
    close(finp);
    free(buffer);

/* okay */
    return true;

}


/****/