/*****************************************************
  file    : c_Link.cpp
  usage   : code definition de c_Link 
  date    : 2001
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  :
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/


#include <c_defs.h>
#include <c_Link.h>



/*-----------------------------------------------------------*/
/*  c_Link
/*-----------------------------------------------------------*/

void c_Link::clear()    // Efface directement l'historique
{
  if( !m_count) return;

  c_Node* pnode;
  while( m_chain)
  {
    pnode  =m_chain;
    m_chain=m_chain->m_prev;
    delete pnode;        // virtual destructor !
  }
  m_count= 0; 
  m_chain= m_root= NULL; 
}

/*-----------------------------------------------------------*/
/*  c_Link
/*-----------------------------------------------------------*/

bool c_Link::erase(c_Node* obj)
  {
    if( !m_count ) return false;

    c_Node* pnode= obj ? obj : m_root;

    // Ajuste root/hist
    if( m_root= obj)
      m_root= obj->m_nextsib;
    if( m_chain= obj)
      m_chain= obj->m_next;


    // Effacement des Childs
    if( pnode->m_child)            // Branche ?
       _erase(pnode->m_child);

    // Chainages selon l'arborescence
    if( pnode->m_nextsib)		        // suivant ?
        pnode->m_nextsib->m_prevsib = pnode->m_prevsib;  

    if( pnode->m_prevsib)		        // prcdent ?
        pnode->m_prevsib->m_nextsib = pnode->m_nextsib;  
    else
      if( pnode->m_parent)          // Premier de liste ?
        pnode->m_parent->m_child = pnode->m_nextsib;  

    // Effacement du node selon la liste
    if( pnode->m_next)
      pnode->m_next->m_prev = pnode->m_prev;  
    if( pnode->m_prev)
      pnode->m_prev->m_next = pnode->m_next;  


    // Dcrmente le nombre d'lments
    if( 0== (-- m_count))
    {
      m_chain= m_root = NULL;
    }

    // Efface le node et retourne
    delete pnode;
    return true;
  }

void c_Link::_erase(c_Node* pnode)
  // Fonction par rcurrence
  // Elimine le node ET TOUS les nodes secondaires de l'objet
  {

      if( pnode->m_child)             // suppression des branches
        _erase(pnode->m_child);

      if( pnode->m_nextsib)		        // suppression des suivants
        _erase( pnode->m_nextsib);

    // Chainages selon l'arborescence
      if( pnode->m_prevsib)		        // Avance ?
          pnode->m_prevsib->m_nextsib = pnode->m_nextsib;  
      else
        if( pnode->m_parent)          // Premier de liste ?
          pnode->m_parent->m_child = pnode->m_nextsib;  
  
      delete pnode;                   // Virtual destructor !!!
      m_count--;              // Dcrmente le nombre d'lments
  }


c_Node* c_Link::_init(c_Node* obj, size_t ukey)     
  // here dfini ou non modifi
  // Retourne NULL si cl duplique
  {
      if(!obj) return NULL; 
      if( !ukey)
        ukey= m_keygen++;                  // JAMAIS 0 ! nouvelle ukey
      else
      {
        if( find(ukey)) return NULL;       // pas de duplicatas
        m_keygen= c_max( m_keygen, ukey);   // ajuster keygen
      }
      if( ! (m_keygen+1)) return NULL;    // Dsol: keygen full!!!

    // Reset des valeurs c_Node
      obj->m_ukey = ukey;
      obj->m_match= true;
      obj->m_parent=obj->m_child= NULL;
      obj->m_nextsib=obj->m_prevsib= NULL;

    // Initialisation historique
    // Chainages de liste/m_chain
      if( !m_count)
        m_chain= m_root= obj; 
      else
      {           
        obj->m_next= NULL;    
        obj->m_prev= m_chain;
        m_chain->m_next= obj;  
        m_chain= obj;           // le dernier
      }

      m_count++;
      return obj;
  }

c_Node* c_Link::find(size_t ukey) const
// Si ukey==0 retourne m_root. (Usage non valide) 
{
  if( !m_count || !ukey) return m_root;

  c_Node* pnode= m_root;  // list contexte toujours

  while( pnode && (pnode->m_ukey != ukey) )
    pnode= pnode->m_next;

  return pnode; 
}

/*-----------------------------------------------------------*/
// Retrouver

size_t  c_Link::count (c_Node* obj) const
// Nombre de descendants ou m_count si NULL
// m_here const
{
  if( !m_count || !obj) return m_count;   // historique
  if( !obj->m_child) return 0;              // pas de child

  obj= obj->m_child;
  size_t  count=1; 
  while( obj->m_nextsib)
  {
    obj= obj->m_nextsib; 
    count++;
  }
  return count;
}

// Retourne le dernier sibling du node parent paramtre
// Si NULL: retourne le dernier node racine
// NOTE: utiliser chain() pour le dernier node cr
c_Node* c_Link::last(c_Node *obj) const
{
  if( !obj)           
    obj= m_root;                      // Utilise la racine
  else
  {
    if( !obj->m_child) return NULL;   // Pas de child!
    obj= obj->m_child;                // liste des childs
  }

  while( obj && obj->m_nextsib)
    obj= obj->m_nextsib; 

  return obj;
}

/*-----------------------------------------------------------*/
// Insertions 

bool c_Link::add    (c_Node* parent, c_Node* obj, size_t ukey) // TreeContext
  {
    if( !_init(obj, ukey)) return false;     // initialisation
    if( 1==m_count) return true;            // rien d'autre  faire

    c_Node* lastsib= last(parent); 

    obj->m_parent = parent;
    if( !lastsib)                     // Creation de child ? 
      parent->m_child = obj;
    else                              // append la list de siblings
    {
      lastsib->m_nextsib = obj;
      obj->m_prevsib= lastsib;
    }

    return true;
  }

bool c_Link::insert     (c_Node* pto, c_Node* obj, size_t ukey)     
  // Inserer devant here
  // here non modifi
  {
      if( !_init(obj, ukey)) return false;     // initialisation
      if( 1==m_count) return true;            // rien d'autre  faire

	  if( pto == obj)	return true;				// C'est le meme
	  if( !pto) 
		pto= m_root;						// sur la racine 

      if( m_root == obj)
        m_root = obj->m_nextsib;

      if( m_root == pto)                    // Ajuster root
        m_root = obj;

    // Chainage arborescent 
      obj->m_prevsib = pto-> m_prevsib;      // prev|<-Obj|Here]
      obj->m_nextsib = pto;                  // prev|Obj->|Here]
      obj->m_parent  = pto->m_parent;        // Parent|<-Obj|(Here)
      
      if( pto-> m_prevsib)
        pto->m_prevsib->m_nextsib = obj;     // prev->|Obj|Here]
      else
        if( pto-> m_parent)
          pto->m_parent->m_child= obj;       // Parent->|Obj|(Here)
        
      pto->m_prevsib = obj;                  // prev|Obj|<-Here]

      return true;
  }


/*
Deplace from dans les siblings et insre devant node(=NULL)
Seuls les chainages d'arborescence sont modifis
Les chainages de liste ne sont pas concerns :)
*/
void c_Link::moveto(c_Node* pto, c_Node* obj)
  {
    if( !obj) return;
	if( pto == obj)	return;				// C'est le meme
	if( !pto) 
	  pto= m_root;						// sur la racine 

    if( pto->m_nextsib==obj) return;

    // correction de root
    if( m_root== obj)         
      m_root= obj->m_nextsib;

    if( m_root== pto)         
      m_root= obj;


    // detach selon arborescence: les childs restent intacts!
    if( obj->m_parent && (!obj->m_prevsib))
        obj->m_parent->m_child = obj->m_nextsib;

    if( obj->m_prevsib)		        // prcdents
        obj->m_prevsib->m_nextsib = obj->m_nextsib;  

    if( obj->m_nextsib)		        // suivants
        obj->m_nextsib->m_prevsib = obj->m_prevsib;  

    // attach selon arborescence 
    obj->m_prevsib = pto-> m_prevsib;      // prev|<-Obj|Here]
    obj->m_nextsib = pto;                  // prev|Obj->|Here]
    obj->m_parent  = pto->m_parent;        // Parent|<-Obj|(Here)
    
    if( pto-> m_prevsib)
      pto->m_prevsib->m_nextsib = obj;     // prev->|Obj|Here]
    else
      if( pto-> m_parent)
        pto->m_parent->m_child= obj;       //Parent->|Obj|(Here)

    pto->m_prevsib = obj;
      
    // correction de root
    if( m_root== pto)
      m_root= obj;
  }

/*
Installe from comme child du node To
Seuls les chainages d'arborescence sont modifis
Les chainages de liste ne sont pas concerns :)
*/
void c_Link::movetochild(c_Node* pto, c_Node* obj)
  {
    if( !obj) return;
	if( pto == obj)	return;				// C'est le meme
	if( !pto) 
	  pto= m_root;						// sur la racine 

    // correction de root
    if( m_root== obj)         
      m_root= obj->m_nextsib;
  
    // detach selon arborescence: les childs restent intacts!
    if( obj->m_parent && (!obj->m_prevsib))
        obj->m_parent->m_child = obj->m_nextsib;

    if( obj->m_prevsib)		        // prcdents
        obj->m_prevsib->m_nextsib = obj->m_nextsib;  

    if( obj->m_nextsib)		        // suivants
        obj->m_nextsib->m_prevsib = obj->m_prevsib;  

    // attache comme child de pto
    c_Node* lastsib= last(pto);
    if( !pto->m_child)
      pto->m_child= obj;
    obj->m_parent  = pto;           // Parent
    obj->m_nextsib = NULL;     
    obj->m_prevsib = lastsib;       // possiblement NULL
    if( lastsib)
      lastsib->m_nextsib= obj;
    
    
  }

/*-----------------------------------------------------------*/
// Parsing de la liste ou de l'arbre 
// Here inchang

bool c_Link::parse(c_Node* pnode, bool (*func)(c_Node* p) )
{
    if( !m_count ) return false;
    if( !pnode )   pnode= m_root;

    for( ; pnode; pnode= pnode->m_nextsib)
    {
      if( false == ((func) ? func(pnode) : eventparse(pnode)))
        return false;

      if( pnode->m_child)                 // recurrence ?
        if( false == parse(pnode->m_child, func))
          return false;
    }
    return true;
}


/*-----------------------------------------------------------*/
// Manipulators

void   c_Link::match   (c_Node* pnode, bool state)
{
  if( !m_count) return;
  if( !pnode) pnode= m_root; 
  pnode->m_match= state;
}

/*--*/

static bool _matchalltrue( c_Node* pnode) 
{ pnode->match(true); return true; }

static bool _matchallfalse( c_Node* pnode) 
{ pnode->match(false); return true; }

void   c_Link::matchall  (c_Node* pnode, bool state)
{
  if( !m_count) return;
  if( !pnode) pnode= m_root; 
  parse( pnode, ((state)?_matchalltrue:_matchallfalse));
}

/*--*/
/*  Renumration
  Il est possible de spcifier newkey < m_keygen
  Break et FALSE si les newkey++ crasent un node possdant la clxistant
*/
bool c_Link::_renum   (c_Node* pnode)
{
  for( ; pnode; pnode= pnode->m_nextsib)
  {
    if( find(m_keygen)) return false;       // D'abord affecter ukey
    pnode->m_ukey= m_keygen++;

    if( pnode->m_child)                     // recurrence ?
      if( false== _renum (pnode->m_child))
        return false;
  }
  return true; 
}

bool c_Link::renum   (c_Node* pnode, size_t newkey)
{
  if( !m_count || !newkey) return false;   // aucun sens, pas de key=0
  if( !pnode) pnode= m_root; 

  size_t savekey= m_keygen;           // push de keygen
  m_keygen= newkey;

  bool ret= _renum(pnode);            // qui modifiera m_keygen

  m_keygen= c_max(m_keygen,savekey);  // la plus haute valeur

  return ret;
}

/*--*/
/* Sort
/*  Sort des siblings: TreeContext toujours
/*  Les chainages de liste ne sont pas concerns
*/

int c_Link::eventsort(const c_Node* p1, const c_Node* p2 )
{ return ( p2->m_ukey - p1->m_ukey ) ; }

bool c_Link::sort(c_Node* pnode, int (*func)(const c_Node* p1, const c_Node* p2) )
{
static  size_t  countSP= 0;       // level de profondeur
static  c_Node* base_sort= NULL;  // vrifier le resync de sort

  if( !m_count ) return false;
  if( !pnode )   pnode= m_root;

  if( !countSP) base_sort= pnode;       // origine du tri

  for( ; pnode; pnode= pnode->m_nextsib)
  {
    if( pnode->m_child)                 // recurrence ?
    {
      countSP++;
      sort(pnode->m_child, func);
      countSP--;
    }

    if(! pnode->m_nextsib) continue;   // et return

    int ret= (func)
            ? func( pnode, pnode->m_nextsib) 
            : eventsort( pnode, pnode->m_nextsib) 
            ;

    if(-1==ret)   // swap des chainages de sibling uniquement
    {
      c_Node* p1 = pnode;
      c_Node* p2 = pnode->m_nextsib;
      c_Node* ac; // accumulateur

      // ajuster parent->m_child si ncessaire
      if( (p1->m_parent) && !(p1->m_prevsib) )
          p1->m_parent->m_child = p2;

      // swap des siblings
      ac=  p2->m_nextsib;  p2->m_nextsib= p1; p1->m_nextsib= ac;
      ac=  p1->m_prevsib;  p1->m_prevsib= p2; p2->m_prevsib= ac;

      pnode= base_sort;   // et recommence
    }
  }

  return true;

}




/*--*/