/*****************************************************
  file    : c_string.cpp
  usage   : chaine extension avec fonctions std::string/stream
  date    : 1991--
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  :
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/

//  Pas de contrle d'erreurs!
//

#include <c_defs.h>
#include <c_string.h>
#include <stdio.h>   // pour ::sprintf()
#include <stdarg.h>   // pour c_str::printf()

#include <crtdbg.h>

#define C_SIZE        (m_end-m_data)
#define C_CHARBLOCK   6            // allou aux 64 bytes suprieurs
/* Allocate
  Blocs de 64 bytes arrondis au bloc suprieur
  Ajoute 8 bytes de scurit
*/
#define C_NEWSIZE(b)  ((1+ ((b)>>C_CHARBLOCK)) << C_CHARBLOCK)			


// static lpstr c_strtok_default= " \t\n";     // sparateurs par defaut

/*----------------------------------------------------*/
/*  c_Str:: primitives
/*----------------------------------------------------*/

bool c_Str::_memalloc(int  bytes)	
// retourne la longueur alloue
{
  // rtablir le contexte de tokening
    toksync();
    _tokraz();

    if( 0== m_mem)					// 1er setup?
    {						   
	    m_mem = C_NEWSIZE(bytes);
	    m_data= (lpstr) malloc( m_mem);
	    _ASSERTE(m_data);
	    memset( m_data, 0, m_mem);
	    m_tokpos= 0;
      m_end= m_data;
      return true;
    }

    int  freesp= m_mem - C_SIZE;   // freespace? 
    if( freesp <= bytes )          // realloc mme si tout just
    {
    int  offset= C_SIZE;           // pour recalculer m_end!!
    int  newmem= C_NEWSIZE(m_mem + bytes - freesp);           
	    m_data = (lpstr) realloc( m_data, newmem);
	    _ASSERTE(m_data);
	    m_mem  = newmem;
      m_tokpos= 0;
      m_end = m_data + offset; 
    }

    return true;
}


void c_Str::_clear() 
{ 
  if( m_tokstr) free (m_tokstr);
  _tokraz();

  if( m_data) free (m_data);  
  m_mem= 0;
  m_data= m_end= NULL; 

} 

/*
  si nLen==0 Prend toute la chaine
  si nLen!=0 Ne prend que les premiers len caractres
*/
bool c_Str::_init(lpcstr sz, int pos, int nLen)
{
  // Initialisation 
    m_mem=0; 
    m_end= m_data= NULL;
    _tokraz();

  // Allocate la taille demandee ou la chaine complte
    if( c_isvalid(sz)) 
    {
      int _len= strlen(sz);
      if( pos < 0 ) pos+= _len;
      if( nLen > _len-pos) 
        nLen = _len-pos;
    }
    else
      nLen= 0;

    _memalloc( nLen+1);      

  // Recopie la chaine
    if( !c_isvalid(sz)) return false;	// Vide ou Null
    c_memcpy( m_data, sz+pos, nLen);

    m_end+= nLen;
    *m_end= 0;
    return true;
}


/*----------------------------------------------------*/
/*  c_Str::c_Str constructors
/*----------------------------------------------------*/

c_Str::c_Str( char  ch, int  count)
{
  _init( NULL, 0, count+1);
  memset( m_data,ch,count);
  m_end= c_adr_nul(m_data);
}


c_Str::c_Str( long val, int base)
{
    _init ( NULL,0,32);             // un block
    if( 10== base) 
      ltoa( val, m_data, base);
    else
      ::sprintf( m_data, "0%0.8X", val);
    m_end= c_adr_nul(m_data);   // attention : ch=0 ?
}



/*  
basic_string::reserve
void reserve(intype n = 0);
The member function ensures that capacity() henceforth returns at least n.
*/
void c_Str::reserve(int bytes)
{
    if( bytes > (int) C_SIZE )
      append( bytes - C_SIZE, 0);
    return;
}

/*
  ensures that size() henceforth returns n. 
  If it must lengthen the controlled sequence, it appends elements with value c.
*/
void c_Str::resize(int bytes, char ch)
{
    // tester incrment
    if( bytes > C_SIZE )
    {
      append( bytes - C_SIZE, ch);
      return;
    }

    // tester si reduction ncessaire
    // Laisse tjrs au moins un bloc
    int newsize= C_NEWSIZE(bytes) ;
	  if( newsize < m_mem )
    {
	    m_mem = C_NEWSIZE(bytes);
	    m_data= (lpstr) realloc( m_data, m_mem);
    }
    m_tokpos= 0;
    m_end = m_data+ bytes;               // rectifie la taille
    *m_end= 0;

}


/*----------------------------------------------------*/
/*  basic_string semantique
/*----------------------------------------------------*/

c_Str& c_Str::append (lpcstr sz, int pos, int  count)
{
    if( !c_isvalid(sz)) return *this;

    int  len= strlen(sz);
    if( pos<0) pos= len-pos;
    len= len-pos;

    if( count > len) count = len;

    _memalloc( count+1 );
    c_memcpy( m_end, sz+pos, count);

    m_end += count;
    *m_end= 0;
    return *this;
}

c_Str& c_Str::append (int  count, char ch)
{
    _memalloc( count);    
    memset  ( m_end, ch, count); 
    if( !ch)
      m_end  = c_adr_nul(m_data);
    else
    {
      m_end += count;
      *m_end = 0;
    }
    return *this;
}

/*
  Operators
*/

c_Str& c_Str::operator= (lpcstr sz)
{
  if( !c_isvalid(sz)) { _clear(); return *this; }
  if( sz== m_data)    return *this;
  int len= strlen( sz);
  resize ( len+1);
  c_memmove( m_data,sz,len);
  m_end= m_data+len;
  *m_end= 0; 
  return *this;
}
c_Str& c_Str::operator= (char ch)
{
  resize ( 2);
  m_data[0]= ch; 
  m_end= m_data+1;
  *m_end= 0; 
  return *this;
}
c_Str& c_Str::operator= (int n)
{
  resize ( 32);
  itoa (n, m_data, 10); 
  m_end= c_adr_nul(m_data);
  return *this;
}
c_Str& c_Str::operator= (long n)
{
  resize ( 32);
  ltoa (n, m_data, 10); 
  m_end= c_adr_nul(m_data);
  return *this;
}

c_Str& c_Str::operator<< (char ch)    // append
{ return append( 1, ch); }

c_Str& c_Str::operator<< (int n)
{
  reserve ( 32);
  itoa (n, m_end, 10); 
  m_end= c_adr_nul(m_data);
  return *this;
}
c_Str& c_Str::operator<< (long n)
{
  reserve ( 64);
  ltoa (n, m_end, 10); 
  m_end= c_adr_nul(m_data);
  return *this;
}

c_Str& c_Str::operator<< (lpcstr sz)   // append lpcstr
{ return append( sz); }

c_Str& c_Str::operator<< (c_Str& sz)   // append c_Str
{ return append( sz); }

/*
  Smart accs
  [>=0] caractre depuis le dbut
  [-1]  caractre depuis la finbut
*/
char c_Str::operator[] (int pos) { return charat(pos); }
char c_Str::charat(int pos)
{
static char nul=0;

  if( pos<0) pos+= C_SIZE;
  if( pos >= C_SIZE) return nul;	// Vide ou Null
  return m_data[pos];     // index positif
}

long c_Str::value(int pos, int len)
  {
  long prmt;
  if( pos < 0) pos+= C_SIZE;
  if( pos >= C_SIZE) return 0L;

  bool ishexa = ('0' == m_data[pos])? true : false;
    sscanf( m_data+pos, (ishexa)?"%X":"%d", &prmt);
	  return prmt;
  }	




/* 
  find
  Retourne la position absolue du caractre  partir de pos ou C_STR_NPOS
*/
int c_Str::find(char ch, int pos) const
{
  if( pos < 0 ) pos+= C_SIZE;   // negatif ?
  if( pos >= C_SIZE) return C_STR_NPOS;	// Vide ou Null
  lpstr retP= strchr( m_data+pos, ch);
  return (retP) ? retP - m_data : C_STR_NPOS; 
}

int c_Str::rfind(char ch, int pos) const
{
  if( pos < 0 ) pos+= C_SIZE;   // negatif ?
  if( pos >= C_SIZE) return C_STR_NPOS;	// Vide ou Null

  lpstr retP= strrchr( m_data+pos, ch);
  return (retP) ? retP - m_data: C_STR_NPOS; 
}

int c_Str::find(lpstr sz, int pos) const
{
  if( !C_SIZE ) return C_STR_NPOS;	// Vide ou Null
  if( pos < 0 ) pos+= C_SIZE;   // negatif ?
  if( pos >= C_SIZE) return C_STR_NPOS;	// Vide ou Null

  lpstr retP= strstr( m_data+pos, sz);                  
  return (retP) ? retP - m_data: C_STR_NPOS; 
}

int c_Str::rfind(lpstr sz, int pos) const
{
  if( !C_SIZE ) return C_STR_NPOS;	// Vide ou Null
  if( pos < 0 ) pos+= C_SIZE;   // negatif ?
  if( pos >= C_SIZE) return C_STR_NPOS;	// Vide ou Null

  lpstr retP, altP;
  for( retP=m_data; altP= strstr( retP, sz); retP= altP)
    ;
  return (retP) ? retP - m_data: C_STR_NPOS; 
}




/*
  Insertion 
*/
c_Str& c_Str::insert(int pos, int count, char ch)
{ 
  if( !ch) return *this;      // tant pis
  if( pos < 0 ) pos+= C_SIZE;   // negatif ?
  if( pos >= C_SIZE) return append( count, ch);

  _memalloc( count);
  lpstr posAdr= m_data+pos;             // adresse d'insertion
  c_memmove( posAdr+count, posAdr, m_end - posAdr);  // Move:: 
  c_memset ( posAdr, ch, count);                     // Insert:: sans NUL
  m_end+= count;
  *m_end= 0;
  return *this;
}

c_Str& c_Str::insert(int pos, lpcstr sz, int pos2, int len2)
{
    if( !c_isvalid(sz)) return *this;
    if( pos < 0 ) pos+= C_SIZE;   // negatif ?
    if( pos >= C_SIZE) return append( sz, pos2, len2);
   
    int count= strlen(sz);        // vrifie longueur
    if( pos2 < 0 ) pos2+= count;
    count-= pos2;

    if( len2 < count ) count= len2; 

    _memalloc( count+1);

    lpstr posAdr= m_data+pos;             // adresse d'insertion

    c_memmove( posAdr+count, posAdr, m_end - posAdr);  // Move:: 
    c_memcpy ( posAdr, sz, count);                     // Insert:: sans NUL

    m_end+= count;
    *m_end= 0;

    return *this;
}

/*
  erase
*/
c_Str& c_Str::erase  (int pos, int  len)
{
    if( !C_SIZE ) return *this;	          // Vide ou Null
    if( pos < 0 ) pos+= C_SIZE;   // negatif ?
    if( pos >= C_SIZE) return *this;	// outbound
    if( len > C_SIZE) len = C_SIZE; 

    lpstr   adr = m_data + pos;         // adresse d'insertion
    int     rest= C_SIZE-pos-len;       // taille du bloc restant
    c_memmove( adr, adr+len, rest);     // overwrite 
    m_end -= len;              
    *m_end= 0;
    _tokraz();

    return *this;
}

/*
  replace
  Depuis startpos sur len caractres maximum, avec sz
  Ajustement de taille si ncessaire
*/

c_Str& c_Str::replace(int pos, int count, lpcstr sz, int pos2, int  count2)
{
lpstr adr1;
lpcstr adr2;
  
  if( !C_SIZE) return *this;
  if( pos < 0 ) pos+= C_SIZE;   // negatif ?
  if( pos >= C_SIZE) return append( sz, pos2, count2); 
  count= c_minor( count, C_SIZE-pos ); 
  adr1= m_data + pos;

  if( !c_isvalid(sz)) return *this;
  int lenSZ= strlen(sz);
  if( pos2 < 0 ) pos2+= lenSZ;

  if( pos2 >= lenSZ) return *this;
  count2 = c_minor ( count2, lenSZ-pos2 );
  adr2= sz + pos2;

  if( count >= count2)  // Espace destination suffisant ?
  {
    c_memcpy ( adr1, sz + pos2, count2);
    c_memmove( adr1 + count2, adr1 + count, (m_end - (adr1 + count)));
    m_end -= (count-count2);
    *m_end = 0;
  }
  else
  {
    erase ( pos, count);
    insert( pos, sz, pos2, count2);
  }

  return *this;
}


/*
The member function returns an object whose controlled sequence is a copy 
of up to n elements of the controlled sequence beginning at position pos.
*/

c_Str c_Str::substr (int pos, int  len) const
{ 
  if( pos < 0 ) pos+= C_SIZE;
  if( pos >= C_SIZE )    return c_Str("");	  // outbound
  if( len > C_SIZE-pos) len = C_SIZE-pos;	            

  return c_Str(m_data,pos,len);   // SUR LA PILE !!
}


/*
Each member function compares up to n0 elements of the controlled sequence 
beginning with position p0, or the entire controlled sequence if these arguments 
are not supplied, to the operand sequence. The function returns:

. A negative value if the first differing element in the controlled sequence 
  compares less than the corresponding element in the operand sequence 
  (as determined by T::compare), or if the two have a common prefix but the 
  operand sequence is longer. 
. Zero if the two compare equal element by element and are the same length. 
. A positive value otherwise
*/

int   c_Str::compare( int pos, int count, lpcstr sz, int pos2, int  count2) const
{
lpstr adr1;
lpcstr adr2;
  if( !C_SIZE)              return -1;
  if( pos < 0) pos+= C_SIZE;   // negatif ?
  if( pos >= C_SIZE)        return -1; 

  if( count > C_SIZE-pos ) count= C_SIZE-pos; 
  adr1= m_data + pos;

  if( !c_isvalid(sz))       return 1;
  int len2= strlen(sz);
  if( pos2 >= len2)         return 1;
  if( pos2 < 0 ) pos2+= len2;   // negatif ?
  if( count2 > len2-pos2 ) count2= len2-pos2;

  // tailles diffrentes ?
  if( count != count2 ) return count - count2;

  adr2= sz + pos2;

  // Boucle
  count= c_minor( count, count2);
  int ret=0; 
  while(!ret && count-- && *adr1 && *adr2)
  {
	  ret= *adr1++ - *adr2++;
  }
  return ret;
}

int   c_Str::icompare( int pos, int count, lpcstr sz, int pos2, int  count2) const
{
lpstr adr1;
lpcstr adr2;
  if( !C_SIZE)          return -1;
  if( pos < 0 ) pos= C_SIZE-pos;   // negatif ?
  if( pos >= C_SIZE )   return -1; 

  if( count > C_SIZE-pos ) count= C_SIZE-pos; 
  adr1= m_data + pos;

  if( !c_isvalid(sz))   return 1;
  int len2= strlen(sz);
  if( pos2 >= len2)     return 1;
  if( pos2 < 0 ) pos2+= len2;   // negatif ?
  if( count2 > len2-pos2 ) count2= len2-pos2;

  // tailles diffrentes ?
  if( count != count2 ) return count - count2;

  adr2= sz + pos2;

  // Boucle
  int ret=0; 
  while(!ret && count--)
  {
	  ret= tolower(*adr1)-tolower(*adr2);
	  adr1++, adr2++;
  }
  return ret;
}


/*
  find_first_not_of
  Trouver squences de chars
  Retourne la position absolue ou C_STR_NPOS
*/
int c_Str::find_first_not_of(char ch, int pos) const
{
    if( !ch)            return C_STR_NPOS;
    if( !C_SIZE)        return C_STR_NPOS;
    if( pos < 0 ) pos+= C_SIZE;   // negatif ?
    if( pos >= C_SIZE)  return C_STR_NPOS;	// outbound
    if( !ch) return C_STR_NPOS;          // Cherche fin de chaine

    lpcstr posAdr= m_data+pos;
    while( *posAdr++ == ch )        // Boucle sur len caractres de chaine
      ; 
    return (*posAdr) ? posAdr-m_data : C_STR_NPOS;
}

int c_Str::find_first_not_of(lpcstr sz, int pos, int  len) const
{
    if( !c_isvalid(sz)) return C_STR_NPOS;
    if( !C_SIZE)        return C_STR_NPOS;
    if( pos < 0 ) pos+= C_SIZE;   // negatif ?
    if( pos >= C_SIZE)  return C_STR_NPOS;	// outbound

    lpcstr posAdr= c_adr_not_of( m_data+pos, sz);
    if( ! *posAdr) return C_STR_NPOS;
    return (posAdr < (m_data+pos+len)) ? posAdr-m_data : C_STR_NPOS;
}

/*
  find_first_not_of
  Trouver squences de chars
  Retourne la position absolue ou C_STR_NPOS
*/
int c_Str::find_last_not_of(char ch, int pos) const
{
    if( !ch) return C_STR_NPOS;   
    if( pos < 0 ) pos+= C_SIZE;   // negatif ?
    if( pos >= C_SIZE) pos = C_SIZE;	

    lpcstr posAdr= m_data+pos;
    while( (posAdr >= m_data) &&  (*posAdr-- == ch) )   
      ; 
    return (posAdr) ? posAdr-m_data : C_STR_NPOS;
}

int c_Str::find_last_not_of(lpcstr sz, int pos, int len) const
{
    if( !c_isvalid(sz))  return C_STR_NPOS;
    if( pos < 0 ) pos+= C_SIZE;   // negatif ?
    if( pos >= C_SIZE) pos= C_SIZE;	

    lpcstr posAdr= m_data+pos;
    while( posAdr >= m_data && (len--))            // Boucle sur len caractres de chaine
    {
      if( !strchr(sz, *posAdr--))
        return posAdr - m_data;
    }
    return C_STR_NPOS;
}

/*
  find_first_of
  Trouver squences de chars
  Retourne la position absolue ou C_STR_NPOS
*/
int c_Str::find_first_of(char ch, int pos) const
{
    if( pos < 0 ) pos+= C_SIZE;   // negatif ?
    if( pos >= C_SIZE) return C_STR_NPOS;	// outbound
    if( !ch) return C_STR_NPOS;          // Cherche fin de chaine

    lpcstr posAdr= m_data+pos;
    while( *posAdr++ != ch )        // Boucle sur len caractres de chaine
      ; 
    return (posAdr) ? posAdr-m_data : C_STR_NPOS;
}

int c_Str::find_first_of(lpcstr sz, int pos, int len) const
{
    if( !c_isvalid(sz)) return C_STR_NPOS;
    if( !C_SIZE )       return C_STR_NPOS;
    if( pos < 0 ) pos+= C_SIZE;   // negatif ?
    if( pos >= C_SIZE)  return C_STR_NPOS;	// outbound

    int retPos= strcspn ( m_data+pos, sz);
    if( retPos > len) return C_STR_NPOS;
    retPos += pos;
    return (retPos < C_SIZE ) ? retPos : C_STR_NPOS;
}

/*
  find_last_of
  Trouver squences de chars
  Retourne la position absolue ou C_STR_NPOS
*/
int c_Str::find_last_of(char ch, int pos) const
{
    if( !ch ) return C_STR_NPOS;          
    if( pos < 0 ) pos+= C_SIZE;   // negatif ?
    if( pos >= C_SIZE) pos = C_SIZE;	// outbound

    lpcstr posAdr= m_data+pos;
    while( posAdr>= m_data && *posAdr-- != ch )        // Boucle sur len caractres de chaine
      ; 
    return (posAdr) ? posAdr-m_data : C_STR_NPOS;
}

int c_Str::find_last_of(lpcstr sz, int pos, int len) const 
{
    if( !c_isvalid(sz)) return C_STR_NPOS;
    if( !C_SIZE )       return C_STR_NPOS;
    if( pos < 0 ) pos+= C_SIZE;   // negatif ?
    if( pos >= C_SIZE) pos = C_SIZE;	// outbound

    lpcstr posAdr= m_data+pos;
    while( (posAdr >= m_data) && (len--))            // Boucle sur len caractres de chaine
    {
      if( !strchr(sz, *posAdr--))
        return posAdr - m_data;
    }
    return C_STR_NPOS;
}


/*----------------------------------------------------*/
/*  extensions
/*----------------------------------------------------*/

/*
  Dpile le dernier caractre de la chaine
*/
char c_Str::pop()	
{
int 	pop;
    if( m_end == m_data) return 0;	  // Vide ou Null
    pop= *--m_end;
    *m_end =0;		    // ecraser dernier char
    return pop; 
}    

lpstr c_Str::pop( int count)	
{
lpstr szpop;
    if( !C_SIZE ) return NULL;	  // Vide ou Null
    m_end -= count;		        // ecraser dernier char
    szpop = strdup( m_end);
    *m_end= 0;
    return szpop; 
}    


/*
  Formatage (ecrase ou append)
  Si false: la chaine est vide
  Si true: la chaine est formate
*/

bool c_Str::printf(int maxlen, lpcstr format, ...)
{ 
va_list     prmt;
int         retval;

    reserve( maxlen);             // c'est l'utilisateur qui dcide 
    va_start( prmt, format);
    retval= _vsnprintf( m_data, maxlen, format, prmt);
    va_end(prmt);
    if(-1==retval) return false ;
    m_end  = m_data+retval; 
    *m_end = 0;
    return true;
}

bool c_Str::appendf(int maxlen, lpcstr format, ...)
{ 
va_list     prmt;
int         retVal;
    _memalloc( maxlen);             // c'est l'utilisateur qui dcide 
    va_start( prmt, format);
    retVal= _vsnprintf( m_end, maxlen, format, prmt);
    va_end(prmt);
    if(-1==retVal) return false ;
    m_end= c_adr_nul(m_end);
    return true;
}


/*
(Java)
  Removes white space from both ends of this string. 
  All characters that have codes less than or equal 
  to '\u0020' (the space character) are considered to be white space. 
*/
int c_Str::trim(int mode)	
// Elimine trailing Spaces: retourne taille
{
    if( !C_SIZE) return 0;	// Vide ou Null

    if( ( mode & C_TRIMRIGHT) && (SPC >= m_end[-1]))   // & 2
    {
      for( m_end--; 
          (m_end>= m_data) && (SPC >= *m_end); 
          *m_end--)
        ;
      *++m_end= NUL;
    }

    if( !C_SIZE) return 0;	// Vide ou Null

    if( (mode & C_TRIMLEFT))   // & 1
    {
      int count= c_pos_no_blank(m_data); 
      if( count)
      {
        c_memmove( m_data, m_data+count, C_SIZE-count);
        m_end -= count; 
        *m_end=NUL;
      }
    }

    return C_SIZE; 
}    

int c_Str::trim(lpcstr szDelims, int mode)
{
    if( !c_isvalid(szDelims)) return trim(mode); 

    if( !C_SIZE) return false;	// Vide ou Null

    if( ( mode & C_TRIMRIGHT))
    {
      for( m_end--; 
          (m_end>= m_data) && strchr(szDelims, *m_end);
          *m_end--)
        ;
      *++m_end= NUL;
    }

    if( !C_SIZE) return 0;	// Vide ou Null

    if( ( mode & C_TRIMLEFT))
    {
      int count= c_pos_not_of(m_data, szDelims); 
      if( count)
      {
        c_memmove( m_data, m_data+count, C_SIZE-count);
        m_end -= count; 
        *m_end=NUL;         
      }
    }

    return C_SIZE;
}


/*
  comme substr, mais extrait la chaine
*/
c_Str  c_Str::extract  (int pos, int  len)  // return object
{
  if( pos < 0 ) pos+= C_SIZE;   // negatif ?
  c_Str retSz= substr( pos, len);
  erase(pos,len);
  return retSz;
}


/*-----------------------------------------------------------*/
/*  Fonctions de tokening 
/*-----------------------------------------------------------*/

/* Retablir la chaine en l'tat et dplacer le pointeur  la suite
/* FALSE si fin de chaine: (retourner NULL)
/* TRUE  si la la longueur de la chaine est correctement restituee
/* (Utilis par tokennext() et tokbalance())
*/
int c_Str::toksync()
{
  if( !C_SIZE) return C_STR_NPOS;    // la chaine est vide!;

  if( C_STR_NPOS==m_tokpos) return C_STR_NPOS;  // Fin de recherche dja dfinie
  if( !m_tokch) return (m_tokpos= C_STR_NPOS);  // Fin de recherche arrive

  m_tokpos= C_SIZE+1;        // position aprs dernier dlimiteur
  *m_end  = m_tokch;         // restaurer delimiteur
  m_tokch = 0;               // rtablir 
  sync();                    // restaurer position de m_end
  return m_tokpos;           // position courante
}

/* Retablir la chaine en l'tat et positionner le pointeur de tokening
*/
int c_Str::toksync(int pos)
{
  toksync();
  if( !C_SIZE) return C_STR_NPOS;     // la chaine est vide!;
  if( pos< C_SIZE) pos+= C_SIZE;      
  if( pos>=C_SIZE) return C_STR_NPOS; // outbound
  return (m_tokpos= pos);
}

/*
  Tokenin
Mode=
  C_DFTTOK=0     defaut mode, "tok1::tok2:" = 3 tokens
, C_WDELIM       append delim "tok1:", ":", tok2:" = 3 tokens
, C_STRTOK       standard     "tok1", tok2" = 2 tokens
*/
lpstr c_Str::tokenstart(lpcstr szDelim, C_TOKMODE mode)
{
  toksync();            // reconstituer la chaine
  m_tokpos= 0;
  if( m_tokstr)          // reset du buffer
  {
    free (m_tokstr);
    m_tokstr= NULL;
  }
  if( !c_isvalid(szDelim))  
    return m_data;

  return _token(szDelim, mode);
}

lpstr c_Str::tokennext(lpcstr szDelim, C_TOKMODE mode)
{  return( C_STR_NPOS!=toksync())? _token(szDelim,mode): NULL;  }


// Traitement proprement dit
lpstr c_Str::_token(lpcstr szDelim, C_TOKMODE mode)
{
int posend; 
  
  if( mode!= C_KEEPTOK)
      m_tokmode= mode;

  // Dfinition des paramtres : tester si SPC definit
  if( c_isvalid(szDelim))  // toute la chaine?
  {
    if( m_tokstr) free( m_tokstr);
    m_tokstr = strdup( szDelim);
    m_tokspc = strchr( m_tokstr, SPC) ? true : false; 
  }

  // Mode STROK ? ignorer tous les delimiteurs 
  if( C_STRTOK== m_tokmode) 
  {
    if( C_STR_NPOS== (m_tokpos= find_first_not_of(m_tokstr,m_tokpos)))
      return NULL;                  // point nul: chaine vide
  }
  else  // ignorer espaces au dbut si ncessaire
  {
    if( m_tokspc) 
    {
      if( C_STR_NPOS== (m_tokpos= find_first_not_of(" \t",m_tokpos)))
        return NULL;                // point nul: chaine vide
    }
  }
  // A partir d'ici m_tokpos point debut de chaine retour
  // trouver delimiteur droite
  if( C_STR_NPOS!= (posend= find_first_of(m_tokstr, m_tokpos)))   
  {
    // Extraire avec/sans delimiteurs (qui peut tre un espace :))
    posend += (C_WDELIM==m_tokmode) ? 1 : 0;
    m_tokch = m_data[posend];         // m_tokch!=0)
    m_data [posend]= 0; 
  }

  sync();                           // resynchroniser !!
  return m_data+m_tokpos;           // offset de chaine
}

// Balance et update du contexte de tokening
lpstr c_Str::tokbalance (lpcstr szSep)
{
int posend;
  if( !c_isvalid(szSep)) return NULL;
  if( C_STR_NPOS== toksync()) return NULL;
  if( C_STR_NPOS== (m_tokpos= find_first_of(szSep, m_tokpos) ))
    return NULL;

  if(C_STR_NPOS!=(posend= balance(szSep, m_tokpos)))
  {
    m_tokpos++;   // pointe APRES balance!!!
    m_tokch= m_data[posend];
    m_data[posend]=0;
    sync();       
    return  m_data+m_tokpos;
  }
  // Pas trouv
  m_tokpos= C_STR_NPOS;
  return NULL;
}




/*-----------------------------------------------------------*/
/*  Fonctions de recherche
/*-----------------------------------------------------------*/


/*
  Recherche nocase
*/
int  c_Str::ifind(lpcstr sz, int pos) const
{
  if( !c_isvalid(sz)) return 0;	// Vide ou Null
  if( pos >= C_SIZE) return 0;	// Vide ou Null

  lpcstr data = m_data+pos;
  lpcstr match= sz;
  while( data= strchr( data, *match))  // cherche premier char
  {
    pos= data - m_data;               // Trouv ? new pos
    bool flag= true;
    while( flag)    
    {
      flag= toupper(*data) != toupper(*match) ;
      data++;
      match++;
    }
    if( !*match)                      // trouv ?
      return pos;                     // oui

    match= sz;                        // non: resync et boucle
  }
  return m_end - m_data;                       // non: fin de chaine
}


int c_Str::findtag( int pos, int len, lpcstr userTag) const
/*  Action:
/*  Dans une chaine contenant des NL intercals,
/*	recherche la premiere ligne commencant par tag.
/*  Les espaces devant tag sont ignoreds
/*
/*      "\n\s+Tag"
/*    
/*  Recherche insensitive 
/*	(Les espaces en dbut de <szTag> sont ignors)
/*
/*  Retour:
/*	pos = pointe SUR userTAG, (mais APRES les espaces initiaux)
/*	      ou C_STR_NPOS si pas trouv
/*
/*  Exemples:
/*	    SeekTag("url:")
/*
*/
{
lpcstr szP    = m_data + pos;
lpcstr szTag  = userTag;
int offset;

    if( !c_isvalid(userTag))  C_STR_NPOS;	// Vide ou Null
    if( !C_SIZE ) return C_STR_NPOS;      // c_Str vide? aucun sens	*/
    if( pos < 0 ) pos+= C_SIZE;           // negatif ?
    if( pos >= C_SIZE) return C_STR_NPOS;   // aucun sens	

    if( len > (int)(m_end-szP)) len= (m_end-szP);  

    szTag= c_adr_no_blank(szTag);                 // pas de blanc dans recherche _!	

    szP  = strchr( szP, NL); 
    while( c_isvalid(szP) && (len > (int) (szP-(m_data+pos)))) 
    {

      szP  = c_adr_no_blank(szP+1);               // Premier char aprs NL
      szTag= userTag;                           // synchronise
      for( offset=0; szP[offset] && szTag[offset]; offset++)
      {
        register c1= toupper(szP[offset]);
        if ( c1 != toupper(szTag[offset]) )
          break; // no match
      }

      if( !szTag[offset])  return szP- m_data;         // offset de la chaine
      szP  = strchr( szP, NL);                  // recommence ligne suivante
    }

    return C_STR_NPOS;                          // pas trouv
}

/*	Balance un char* d'un separateur a l'autre, ou jusk'a NUL
/*	Les escapes sont admis
/*  Entree:
/*	  position du caractre  balancer 
/*        Si la position est un blanc, recherche premier !blanc
/*  Retour:		
/*	  adresse du caractre correspondant
/*  Notes: 
/*  . Si le premier caractre rencontr n'est pas une balance, 
/*    alors le tokening retourne le premier espace
/*  . Reconnat les escapes
/*  . Pas d'imbrication de balances
*/
static  lpstr szBALANCE_SEPS= "({<[" ")}>]?!";	    // = eostr() si pas trouved
#define	OFFSET_BALANCE_SEPS	6			    // Offset pour balance
int c_Str::balance (lpcstr szSep, int pos)
{
char  cSep; // Caractere a balancer
lpstr psz;

  if( !c_isvalid(szSep)) return C_STR_NPOS;	// Vide ou Null
  if( !c_isvalid(m_data)) return C_STR_NPOS;	// Vide ou Null
  if( pos < 0) pos+= C_SIZE;	  
  if( pos >= C_SIZE) return C_STR_NPOS;	// Vide ou Null

  if( C_STR_NPOS== (pos= find_first_of(szSep,pos)))   // trouver caractre
    return C_STR_NPOS;	// eostring

// Determiner char a balancer
  psz= strchr( szBALANCE_SEPS, charat(pos));		// = NULL si pas trouved

  if( psz) 
    cSep= psz[OFFSET_BALANCE_SEPS];		        // charactre balanc?
  else 
    if( strchr("\"'/_*", m_data[pos]))          // idem-char (QUOTES/DQUOTES)
      cSep= charat(pos);				                
    else                                      // sinon: chercher espace !
      cSep= SPC;                              

// Cherche la balance: admet char escapeds
  if(SPC!=cSep) 
    pos++; 

  for( psz= m_data+pos; 
      *psz && (cSep!=*psz); 
       psz+= (BSLASH==*psz) ? 2:1)
  ;		

// Retourne position du caractre 
  return ( *psz) ? (psz-m_data) : C_STR_NPOS;	
}
#undef OFFSET_BALANCE_SEPS	// Offset pour balance


////////////////////////////////////////////////////////////////////////
//  HTML conversions
////////////////////////////////////////////////////////////////////////

static struct	
{
    short   m_NumRef;
    char    m_Symbole;
    lpstr   m_Entity;
}
HtmlEntities[]=	    // Note: case signifiant!
{
  0  , '<',   "&lt;"	// LesserThan
, 0  , '>',   "&gt;"	// GreaterThan
, 0  , '&',   "&amp;"	// Ampersand
, 0  ,'\"',   "&quot;"	// DoubleQuotes
,	146, '',   "&apos;"	// English possessive  

,	146, '',   "&#146;"	// English possessive  
,	34, '\"',  "&#34;"	// English possessive  


, 192, '',   "&Aacute;"	// Capital A,  acute accent
,	193, '',   "&Agrave;"	// Capital A,  grave accent
,	194, '',   "&Acirc;"	// Capital A,  circumflex accent
,	195,  0 ,   "&Atilde;"	// Capital A,  tilde
,	196,  0 ,   "&Aring;"	// Capital A,  ring
,	197, '',   "&Auml;"	// Capital A,  dieresis or umlaut mark
,	198,  0 ,   "&AElig;"	// Capital AE dipthong (ligature)
,	199, '',   "&Ccedil;"	// Capital C,  cedilla
,	200, '',   "&Eacute;"	// Capital E,  acute accent
,	201, '',   "&Egrave;"	// Capital E,  grave accent
,	202, '',   "&Ecirc;"	// Capital E,  circumflex accent
,	203, '',   "&Euml;"	// Capital E,  dieresis or umlaut mark
,	204, '',   "&Iacute;"	// Capital I,  acute accent
,	205, '',   "&Igrave;"	// Capital I,  grave accent
,	206, '',   "&Icirc;"	// Capital I,  circumflex accent
,	207, '',   "&Iuml;"	// Capital I,  dieresis or umlaut mark
,	208,  0 ,   "&ETH;"	// Capital Eth,  Icelandic
,	209, '',   "&Ntilde;"	// Capital N,  tilde
,	210, '',   "&Oacute;"	// Capital O,  acute accent
,	211, '',   "&Ograve;"	// Capital O,  grave accent
,	212, '',   "&Ocirc;"	// Capital O,  circumflex accent
,	213,  0 ,   "&Otilde;"	// Capital O,  tilde
,	214, '',   "&Ouml;"	// Capital O,  dieresis or umlaut mark
 
,	216,  0 ,   "&Oslash;"	// Capital O,  slash
,	217, '',   "&Uacute;"	// Capital U,  acute accent
,	218, '',   "&Ugrave;"	// Capital U,  grave accent
,	219, '',   "&Ucirc;"	// Capital U,  circumflex accent
,	220, '',   "&Uuml;"	// Capital U,  dieresis or umlaut mark;
,	221,  0 ,   "&Yacute;"	// Capital Y,  acute accent
,	222,  0 ,   "&THORN;"	// Capital THORN,  Icelandic
,	223,  0 ,   "&szlig;"	// Small sharp s,  German (sz ligature)
 
,	224, '',   "&aacute;"	// Small a,  acute accent
,	225, '',   "&agrave;"	// Small a,  grave accent
,	226, '',   "&acirc;"	// Small a,  circumflex accent
,	227,  0 ,   "&atilde;"	// Small a,  tilde
,	228,  0 ,   "&aring;"	// Small a,  ring
,	229, '',   "&auml;"	// Small a,  dieresis or umlaut mark
,	230,  0 ,   "&aelig;"	// Small ae dipthong (ligature
,	231, '',   "&ccedil;"	// Small c,  cedilla
,	232, '',   "&eacute;"	// Small e,  acute accent
,	233, '',   "&egrave;"	// Small e,  grave accent
,	234, '',   "&ecirc;"	// Small e,  circumflex accent
,	235, '',   "&euml;"	// Small e,  dieresis or umlaut mark
,	236, '',   "&iacute;"	// Small i,  acute accent
,	237, '',   "&igrave;"	// Small i,  grave accent
,	238, '',   "&icirc;"	// Small i,  circumflex accent
,	239, '',   "&iuml;"	// Small i,  dieresis or umlaut mark
,	240,  0 ,   "&eth;"	// Small eth,  Icelandic
,	241, '',   "&ntilde;"	// Small n,  tilde
,	242, '',   "&oacute;"	// Small o,  acute accent
,	243, '',   "&ograve;"	// Small o,  grave accent
,	244, '',   "&ocirc;"	// Small o,  circumflex accent
,	245,  0 ,   "&otilde;"	// Small o,  tilde
,	246, '',   "&ouml;"	// Small o,  dieresis or umlaut mark
 
,	248,  0 ,   "&oslash;"	// Small o,  slash
,	249, '',   "&uacute;"	// Small u,  acute accent
,	250, '',   "&ugrave;"	// Small u,  grave accent
,	251, '',   "&ucirc;"	// Small u,  circumflex accent
,	252, '',   "&uuml;"	// Small u,  dieresis or umlaut mark
,	253,  0 ,   "&yacute;"	// Small y,  acute accent
,	254,  0 ,   "&thorn;"	// Small thorn,  Icelandic
,	255,  0 ,   "&yuml;"	// Small y,  dieresis or umlaut mark
 
,	191, '',  "&#191;"	// inverted question mark
,	161, '',  "&#161;"	// inverted exclamation mark    

,	174,  0 ,   "&reg;"	// Registered    
,	169,  0 ,   "&copy;"	// Copyright

#if 0		// inconnus!
,	0  ,  0 ,   "&trade;"	// TradeMark
,	0  ,  0 ,   "&nbsp;"	// Non breaking space
#endif

,	-1,   0 ,   0 
};


void c_Str::htmldecode()
/****************************************/
/*  Conversion d'entites HTML en caracteres ISO-8859-1
*/
{
int pos, end;
short loop;

    if( !C_SIZE) return; 

    pos=0;
    while( C_STR_NPOS!= find(AMPERSAND,pos))
    {
      end= find( SEMICOLON, pos+1);
      if( DIESE== m_data[pos+1])
      {
        replace(pos, end-pos, atoi(m_data+pos) ); 
      }
      else
      {
      for( loop= 0 ; -1!= HtmlEntities[loop].m_NumRef; loop++)
	      if( compare( HtmlEntities[loop].m_Entity))
        {
	          replace(pos, end-pos, HtmlEntities[loop].m_Symbole);	    
            break;
        }
      }
      pos++;    // un seul caractre!
    }
}


void c_Str::htmlencode( int mode)
/****************************************/
/*  Conversion de ISO-8859-1 en entites HTML 
/*  Flags bitmap:
/*  mode= 0  Uniquement GT-LT-AMPERSAND (pour xml ISO)
/*  mode= 1  Packing des NL et espaces rpts
/*  mode= 2  Utilise aliases d'entits
/*  mode= 4  Utilise entits numriques (XML)
*/
{
char lastch;
c_Str swap; 

    if( !C_SIZE) return; 

    swap.reserve ( size()/8*10 );   // allocation d'espace
    for( m_end= m_data; *m_end; m_end++)
    {
      char ch= *m_end;
      switch(ch)
      {
      case AMPERSAND: swap << "&amp;"; continue;
      case LT: swap << "&lt;"; continue;
      case GT: swap << "&gt;"; continue;
      case APOS: swap << "&apos;"; continue;
      case QUOT: swap << "&quot;"; continue;
      case NL: 
        if( ( mode & C_HTMLPACK) && (lastch == ch)  )
          continue; 
        break;
      default:
        if( ch <= SPC) 
        {
          ch = SPC; 
          if( ( mode & C_HTMLPACK) && (lastch == ch)  )
            continue; 
          break;
        }
        if( ( mode &  C_HTMLALIAS|C_HTMLSTRICT) && ( ch > 0x7E)  )
        {

          if( ( mode & C_HTMLALIAS))
          {
            bool flag= false;
            for( int loop=0; -1!=HtmlEntities[loop].m_NumRef; loop++)
            {
    	        if( ch== HtmlEntities[loop].m_Symbole)
              {
	              swap<< HtmlEntities[loop].m_Entity;	    // Retourne literal
                flag= true; 
                break;
              }
            }
            if ( true== flag) continue; 
          }

        // HTMLSTRICT only ou pas d'alias
          swap<< AMPERSAND << (int)ch << SEMICOLON; 
          continue;
        }
      }

      // Traitement par dfaut du caractre
      swap<< (lastch=ch); 
    }

    // m_end gale 0
    operator= (swap.c_str()); 
    return;
}


/*----------------------------------------------------*/
/*  specifics
/*----------------------------------------------------*/
void  c_Str::strip () 
/*  Usage:
/*	    Retourne la MEME chaine SANS espaces dupliqus
/*      - (SPC)+    -> SPC
/*      - SPC+NL+SPC -> NL
/*      - Extraneous blanks respectss 
/*  Obs:
/*    Utilis pour stripping de texte XML
/*    Si szDelim contient SPC, alors SPC/TAB doubles ramens  SPC seul
/*    mais les autres chars<SPC sont prservs
/*	La chaine est reecrite, puisque nSSRment <= en longueur
*/
{
lpstr	psz = m_data;
lpstr	pto = m_data;
char  ch;

  //psz= c_adr_no_blank(psz);         // ltrim()
	while( (ch= *psz++))
	{
	    switch(ch)
      {
      case TAB: 
      case SPC: 
        psz= c_adr_no_space(psz);   // ignorer SPC TAB
        //if( !*psz)  break;          // chop()
        if( NL!= *psz)
        {
          *pto++ = SPC;             // copier SPC
          break;                    // Si NL, assume rgle de NL
        }                           

      case NL: 
        psz= c_adr_no_blank(psz);  // ignorer SPC NL TAB
        //if( !*psz)  break;          // chop()
        *pto++ = NL;                // copier NL 
        break;

      default:
        *pto++ = ch;         // copier le caractres
      }
  }
  *pto= 0;
  m_end= pto; 
}

void c_Str::addslashes()
{
int   pos=0;
lpstr escchar= "\\_";

  while( C_STR_NPOS!= (pos= find_first_of("\"'", pos)))
  {

    escchar[1]= charat(pos);
    replace(pos,1,escchar,2);
    pos+=2;
  }
}





#undef C_SIZE 