/*****************************************************
  file    : c_xem.cpp
  usage   : fonctions entre/sortie pour c_Xem
  date    : 2001
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  :
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/

#include  <c_xem.h>

/*----------------------------------------------------------*/
/*  c_Xemnode
/*----------------------------------------------------------*/

bool  c_Xemnode::toxem(c_Str& buf)
{
  // traitement des balises d'ouverture
  int loop;
  switch( m_type)
  {
  case CX_PI_CMD:     
    buf<< "\n<?" << tagname();
    for( loop=0; loop< countatt(); loop++)
      buf<< SPC << attname(loop) << "=\"" << attdata(loop) << DQUOTES;
    buf<< "?>";
    return true;      // Attention, assume root position 

  case CX_COMMENT:
    buf<< "\n<!--\n" << m_cdata << "\n-->\n";
    break; 

  case CX_CDATA:
    buf<< "\n<CDATA><![" << tagname() << "[" << m_cdata << "]]></CDATA>";
    break; 

  case CX_TEXT:
    buf<< xemstrip(m_cdata);              // concatnation ventuelle des lignes
    break;

  case CX_TAGEND:
    buf<< "<" << tagname();
    for( loop=0; loop< countatt(); loop++)
    {
      buf<< SPC << attname(loop) << "=\"" << attdata(loop) << DQUOTES;
    }
    buf<< "/>"; 
    break;

  case CX_ELEMENT:
    buf<< "\n<" << tagname();
    for( loop=0; loop< countatt(); loop++)
    {
      buf<< SPC << attname(loop) << "=\"" << attdata(loop) << DQUOTES;
    }
    // Traitement des codes de fermeture
    if(!child())
      buf<< "/>"; 
    else
      buf<< ">"; 
    break;

  }

  /* La syntaxe de XEM diffre de XML:
  /* par le fait que les balises de dernier niveau ne
  /* sont pas fermes.
  */
  if( child()) return true;      // continuer
  if( nextsibling()) return true;

  // Boucle de fermeture des parent/idem si le dernier sibling
  c_Xemnode* pnode= CAST_XEMNODE(parent());
  if( !pnode) pnode= this;

  while( !pnode->nextsibling() && (NULL!=pnode->parent()))
  {
      pnode= CAST_XEMNODE(pnode->parent());
      buf<< "</" << pnode->tagname() << ">"; 
  }
  return true;
}
  
bool  c_Xemnode::toxml(c_Str& buf)   // conversion  XML
{
  // traitement des balises d'ouverture
  int loop;
  switch( m_type)
  {
  case CX_PI_CMD:     
    buf<< "\n<?" << tagname();
    for( loop=0; loop< countatt(); loop++)
      buf<< SPC << attname(loop) << "=\"" << attdata(loop) << DQUOTES;
    buf<< "?>";
    return true;      // Attention, assume root position 

  case CX_COMMENT:
    buf<< "\n<!--\n" << m_cdata << "\n-->\n";
    break; 

  case CX_CDATA:
    buf<< "\n<CDATA><![" << tagname() << "[" << m_cdata << "]]></CDATA>";
    break; 

  case CX_TEXT:
    xemstrip(m_cdata);              // concatnation ventuelle des lignes
    buf<< strtoxml(m_cdata);
    break;

  case CX_TAGEND:
    buf<< "<" << tagname();
    for( loop=0; loop< countatt(); loop++)
    {
      buf<< SPC << attname(loop) << "=\"" << attdata(loop) << DQUOTES;
    }
    buf<< "/>"; 
    break;

  case CX_ELEMENT:
    buf<< "\n<" << tagname();
    for( loop=0; loop< countatt(); loop++)
    {
      buf<< SPC << attname(loop) << "=\"" << attdata(loop) << DQUOTES;
    }
    // Traitement des codes de fermeture
    if(!child())
      buf<< "/>"; 
    else
      buf<< ">"; 
    break;

  }

  // Traitement des codes de fermeture
  if( child()) return true;      // continuer

  // Boucle de fermeture des parent/idem si le dernier sibling
  c_Xemnode* pnode= this;
  while( !pnode->nextsibling() && (NULL!=pnode->parent()))
  {
      pnode= CAST_XEMNODE(pnode->parent());
      buf<< "</" << pnode->tagname() << ">"; 
  }
  return true;
}

/*----------------------------------------------------------*/
/*  Traitement des donnees de texte
/*----------------------------------------------------------*/
// (variante de c_String::strip avec concatnation de lignes)
// Entre:
//    la chaine comprise entre deux balises, NL comprises.
// Traitement des espaces
//  - Les NL extrieurs sont supprims (mais les SPC sont conservs)
//  - Les SPC multiples sont ramens  un seul
//  - Chaines de plusieurs SPC|NL sont ramenes  NL ou SPC selon 
//    concatnation
// Sont concatnes:
// - les lignes commenant par minuscules, si la ligne prcdente 
// ne finit pas par une ponctuation ou est elle-meme vide


c_Str& c_Xemnode::xemstrip(c_Str& buf)
{
lpstr pbuf, psz, szbeg;
char ch_toconcat= NL;

  pbuf= buf;
  for( szbeg= pbuf; psz = c_adr_eol(szbeg); szbeg= psz+1)
  {
    bool  needconcat= false;
    lpstr prevpos, postpos;

    prevpos= postpos= psz;                // pointe NL
    while( SPC== prevpos[-1]) prevpos--;  // Premier espace  gauche de NL
    while( SPC== postpos[+1]) postpos++;  // Dernier espace  droite de NL

    ch_toconcat= NL;
    needconcat= false;

    if( PERCENT==postpos[1]) needconcat= true;  // Spcial URL tronqus
    if( iswalpha(postpos[1])&& iswlower(postpos[1])) needconcat= true;
    if( strchr( "\"'/()", postpos[1])) needconcat= true;
    if( strchr( ":?!.\n", prevpos[-1])) needconcat= false;

    if(true==needconcat)
        ch_toconcat= SPC;  

    if( postpos-prevpos)      // il faut liminer des espaces ?
      buf.replace( prevpos-pbuf, postpos-prevpos+1, 1, ch_toconcat);
    else
      *psz= ch_toconcat;       // simplement le sparateur

    psz= prevpos;             // boucler
  }
  buf.sync();
  return buf;
}

/*
Conversions specifiques HTML 
  \n -- <BR/>	   (sauf si \n+ minuscule)
  \{texte=url} --  <a href='url'>texte</a>
  \{texte=#url} -- <a href='#url'>texte</a>
  \n#name:  --     <a name='name'/>
*/
c_Str& c_Xemnode::strtoxml(c_Str& cdata)
{
lpstr szp= cdata;
c_Str szreplace;
c_Str szText;
c_Str szHref;

int   beg, end,pos=0;

  // Remplace NL para <P/> 
  pos=0; 
  while(C_STR_NPOS != (pos= cdata.find(NL,pos)))
  {
    cdata.replace(pos,1,"\n<br/>");
    pos+=5;
  }

  // Formatage des rfrences
  pos=0;
  while(C_STR_NPOS != (pos=cdata.find("\\{", pos)))
  {
    if( C_STR_NPOS == (end= cdata.balance("{", pos+1))) break; // ignorer
    // Degager les tokens
    szreplace= cdata.substr(pos+2, end-pos-2);
    if( C_STR_NPOS == (beg= szreplace.find("::"))) // ignorer
    {
      szreplace.trim();
      szText= szHref= szreplace;
    }
    else
    {
      szText= szreplace.substr(0,beg); 
      szHref= szreplace.substr(beg+2); 
      szText.trim();
      szHref.trim();
    }

    // Dterminer le tag applicable
    szreplace.erase();
    szreplace << "<a href='" << szHref << "'>"
              << szText	  // texte
              << "</a>";

    cdata.replace(pos, end-pos+1, szreplace);
    pos+= szreplace.length();
  }

  // Formatage des anchors
  pos=0;
  while(C_STR_NPOS != (pos=cdata.find("\n#", pos)))
  {
    if( C_STR_NPOS == (end= cdata.find(pos+2,COLON))) break; // Deux points
    szreplace.erase();
    szreplace << "<a name='" << cdata.substr(pos+2, end-pos-2) << "'/>";
    cdata.replace(pos, end-pos+1, szreplace);
  }
  return cdata;
}

/*----------------------------------------------------------*/
/*  c_Xem private
/*----------------------------------------------------------*/

// retrouver premier node >= ELEMENT, TAGEND, COMMENT, CDATA
c_Xemnode*    c_Xem::firstelement()   
{

    c_Xemnode* pnode= CAST_XEMNODE(root());
    
    while( pnode && ( pnode->type() < CX_ELEMENT ))
    {
      pnode= CAST_XEMNODE(pnode->nextsibling());
    }
    return pnode;
}


/*----------------------------------------------------------*/
/*  c_Xem public
/*----------------------------------------------------------*/

bool c_Xem::load(lpstr szFilename)
{

// Chargement du fichier
  m_fp.close();
  if( false== m_fp.open( szFilename))
    throw err( CX_ERR_FILE, "Fichier n'existe pas.");

  _parsedoc();

  m_fp.close();


// Ajout d'un noeud XEM racine, si ncessaire
// 
  c_Xemnode* pnode= firstelement();   // Premier node de type ELEMENT
  if( pnode->nextsibling()) 
  {
    if( ! (pnode= new c_Xemnode()))   // creer node XEM
      throw err( CX_ERR_MEMORY, "memoire insuffisante");

    c_Link::insert(firstelement(), pnode);    // sur la racine     
    pnode->m_type    = CX_ELEMENT; 
    pnode->m_tagname = "xem";                 // tag en clair
	pnode->m_attname << "id"; 
	pnode->m_attdata << "ID0";				  // toujours

    // passe tout les noeuds comme child
    while( pnode->nextsibling())      // deplace les noeuds sous (pnode)
      c_Link::movetochild(pnode, pnode->nextsibling()); 
  }

  return true;
}

/** Ajouter un fichier source au fichier destination existant.
	La racine du fichier destination est conserve et restituee.
	Tous les noeuds (mono- ou multi-racine) du f-source passent
	comme enfants de la racine dj existante du f-destination

*/
bool c_Xem::append(lpstr szFilename)
{
// Utiliser load() si l'arbre est vide
  if( 0== c_Link::count()) 
	return load( szFilename);

c_Xemnode *pnode;			// pointeur sur objet racine

// Examiner s'il est ncessaire de grer la racine unique
  pnode	 = firstelement();
  if( NULL!= pnode->nextsibling())	  // arbre multiracines ?
	return load(szFilename);		  // oui ! pas besoin de grer la racine

// Faire passer tous les childs comme prev-siblings de la racine
  while( pnode->child())
	moveto(pnode, pnode->child());	  // L'insertion se fait *devant* pnode

// La racine est maintenant passe en queue de liste: 
// la remettre en premire position 
  moveto( firstelement(), pnode);

// Lecture normale du fichier  
  m_fp.close();
  if( false== m_fp.open( szFilename))
    throw err( CX_ERR_FILE, "Fichier n'existe pas.");

  _parsedoc();
  m_fp.close();

// Reconstruit l'arbre pour le faire dpendre de pnode
// ASSERT( pnode==firstelement());

  while( pnode->nextsibling())      // deplace les noeuds sous (pnode)
    c_Link::movetochild(pnode, pnode->nextsibling()); 

// Voil.
  return true;
}


bool c_Xem::save( lpstr filename, c_Xemnode* pnode)
{
  m_outnode= pnode;		// node racine 
  m_isoutfile= true;	// sortie vers fichier
  m_outstr = NULL;		

  m_fp.close();
  if( false== m_fp.open( filename, 'w'))
    throw err( CX_ERR_FILE, "Fichier sortie ne peut tre ouvert.");

  // Heading TYPES
  if( CX_XMLTYPE==m_outtype)
    m_fp.putsnl("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>");

  m_fp.putsnl("<?xem version=\"0.1\" licence=\"www.in3activa.org/doc/es/LPT-ES.html\"?>");

  c_Link::parse(pnode);       // invoque c_Xem::eventparse()

  m_fp.putch(NL);             // forcer NL  la fin
  m_fp.close();
  return true;
}

void  c_Xem::string( c_Str& buffer,  c_Xemnode* pnode)
{
  m_outnode= pnode;
  m_isoutfile= false;
  m_outstr = &buffer;

  m_outstr->erase();
  c_Link::parse(pnode);     // invoque c_Xem::eventparse()

  // Tag de fermeture
  if( pnode && !pnode->child() && CX_PI_CMD!=pnode->type())
    buffer << "</" << pnode->tagname() << ">\n";

}


/*----------------------------------------------------------*/
/*  c_Xem eventparse: 
/*----------------------------------------------------------*/

bool c_Xem::eventparse(c_Node* p)     // virtual c_Xem::eventparse()
{
c_Xemnode* pnode= CAST_XEMNODE(p);
c_Str buf;    // buffer transitoire pour outfile

  buf.erase();
  switch( m_outtype)
  {
  case CX_XEMTYPE:    // sortie de type XEM
    if( false== pnode->toxem(buf))
      return false; 
    break;

  case CX_XMLTYPE:    // sortie de type XML
    if( false== pnode->toxml(buf))
      return false; 
    break;
  }

// sortie;
  if( m_isoutfile)
    m_fp<< buf;
  else
    m_outstr->append(buf);
  return true; 
}


