/*****************************************************
  file    : c_xemparse.cpp
  usage   : analyseur de c_Xem
  date    : 1991--
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  :
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/


#include  <c_Xem.h>


/*----------------------------------------------------------*/
/*  Parsing de document  
/*  _parsedoc	lecture du fichier entree   
/*  _eventdata  installe un pseudo-noeud CX_TEXT de donnes accumules
/*  _eventatt   traitement des attb dans balise ouvrante
/*  _eventdoc	traitement des balises elles-mme
/*----------------------------------------------------------*/


int c_Xem::_parsedoc()
{
/**/
c_Str balise; 					// balise en entre
c_Str tagname;					// tag de la balise
c_Str cdata;					// donnes balises
char ch;						// caractre 

/* Note sur imbrications:
Les balises des blocs retenues sont soit:
- spcifie par m_usertag
- dtermines automatiquement par la premire balise rencontre,
  qui est conservee dans tagbloc


La valeur de tagbloc determine par consquent le contexte
hors-bloc ou intra-bloc. Toute donnes hors-bloc est ignore.

Il n'est pas possible d'imbriquer des balises de bloc.
L'apparition intra-bloc d'une balise identifie comme ouvrant un 
bloc permet de dtecter que le bloc prcdent n'a pas t referm.
(Dans ce cas, les donnes accumules jusque l sont ignores)

*/
c_Str tagbloc;					// repre de balise initiale
int	  code;						// caractre lu

CX_TYPE  state=CX_TYPEUNDEF;    // CX_ELEMENT, ATTRIBUTE etc..
bool     implicite=false;		// si syntaxe fermee <../>       
bool     ltspace  =false;		// extraneous blank 
bool     rtspace  =false;		// extraneous blank 

/**/

// verifier que m_basids > 0
// ID0 est rserv au noeud racine
  if( !m_baseids) m_baseids++;  


// boucle de lecture de l'arborescence
  while( !m_fp.eof() )
  {
	rtspace=ltspace=false;				  // pour traitement des espaces
	state=CX_TYPEUNDEF; 

	// Lecture de texte jusqu' la premire balise qui modifie state
	if( tagbloc.empty())				// Hors-bloc ?
	{
	  if( false == (m_fp.seekstr("<"))) continue;
	}
	else
	{
	  if( false == (m_fp.seekstr("<", cdata))) continue;
	}

	// Casuel d'analyse de la balise
	balise.erase();                   
	switch( m_fp.peek())
	{
	case NL:                  // syntaxe balise tronque!
	  m_fp.getch();           // ignorer /</
	  cdata << "&lt;";             
	  break; 
    
	case LT:                  // syntaxe << 
	  m_fp.getch();           // ignorer /</
	  cdata << "&lt;" ;
	  break;               // chercher encore du texte

	case PERCENT:     // PI_CMD mais <% aspcode %>--> CX_CDATA
	  //_ASSERTE("debug"==0);
	  if( false== m_fp.seekstr("%>", balise)) break;

	  tagname = "CDATA";
	  balise.insert(0,1,LT);  // reconstruire la balise
	  balise << "%>";         // (non packed)
	  state   = CX_CDATA;     // le paramtre est la balise 
	  break;

	case QUESTION:    // PI_CMD mais <?php data ?>--> CX_CDATA

	  if( false== m_fp.seekstr("?>", balise)) break;

	  if( SPC>= balise[1])    // balise <? isole : alors <![CDATA[<?php]]>
	  {
		tagname = "CDATA";
		balise.replace(0,1,"<?php");  // remplace ? initial par '<?php'
		balise << "?>";         // (non packed)
		state   = CX_CDATA;     // le paramtre est la balise 
		break;
	  }

	  tagname = balise.tokenstart("? ", C_STRTOK);
	  tagname.lower();                        
  
	  if( c_strequ("xem",tagname)) break;    // ignorer  (output automatique)
	  if( c_strequ("xml",tagname)) break;    // ignorer  (output automatique)
	  /* (mais par ex. xml:stylesheet passera OK ) */

	  if( tagname=="php")     // change  <![CDATA[<?php code ?>]]>
	  {
		balise.toksync();           // annuler tokening;
		tagname= "CDATA";
		balise.insert(0,1,LT);  // reconstruire la balise
		balise << "?>";
		state   = CX_CDATA;	  // le paramtre est la balise 
		break;
	  }

	  state=CX_PI_CMD;        // PI_CMD valide
	  break;

	case EXCLAMATION:         // <!-- (donnes) --> ou <![CDATA[(donnes)]]>

	  m_fp.getch();           // ignorer /!/

	  switch( code= m_fp.getch())
	  {
	  case '-':				  // <!--	-->
		if( false== m_fp.seekstr("-->", balise)) break;
		state=CX_COMMENT;       
		break;

	  case '[':				  // <![CDATA[  ]]>
		tagname.erase();

		if( false== m_fp.seekstr("[", tagname))    // Lecture du TAG
		  break;

		if( false== m_fp.seekstr("]]>", balise))   // Lecture des donnes
		  break;

		state   = CX_CDATA;   // le paramtre est la balise 
		break;

	  default:	  // syntaxe incorrecte: resync
		cdata << "&lt;" << code;
	  }

	  break;

	default:            // Tous nodes CX_ELEMENT ouvrants et fermants
	  /* Notes
		Mme si l'entre est XML, des restrictions s'appliquent:
		- une balise CX_ELEMENT doit figurer complte sur la mme ligne
		*/

	  m_fp.savepos();

	  if( false== m_fp.seekstr(">", balise))	// lire le reste de la balise
		break;

	  // tester si lment final avant de supprimer FSLASH ou Question
	  state= ( FSLASH == balise[0]) 
		   ? CX_ENDBLOCK
		   : CX_ELEMENT;   
	  balise.trim("/", C_TRIMLEFT);		// Noter que <http://www.adr.com> est conserv tel quel

								 // Test nSSR si texte dbute par <qlq/> ???
	  if( FSLASH == balise[-1])  // <tag/> ELEMENT implicite
	  {
		state= CX_TAGEND;
		balise.trim  ("/", C_TRIMRIGHT);
	  }

	  // tester si contient quelque chose
	  balise.strip();			// compression SP+SP= SP, SP+NL= NL
	  if( balise.empty() || C_STR_NPOS== balise.find_first_not_of(" \n"))						// aucun sens
	  { 
		m_fp.lastpos();
		cdata << "&lt;" ;				// conserver comme escape
		state= CX_TYPEUNDEF;					// continuer la lecture 
		break;
	  }

	  // recuperer tagname
	  tagname = balise.tokenstart(" ", C_STRTOK);				   
	  tagname.lower();                        

	  // vrifier que le tagname est OK
	  lpcstr psz= tagname; 
	  bool   isvalide= (isalpha(*psz++)) ? true : false;	// 1er char ascii
	  while( (true== isvalide)  && (ch = *psz++))
	  {
		isvalide= 
		  (isalnum(ch) 
		  || UNDERSCORE==ch 
		  || TIRET==ch 
		  || COLON==ch)
		  ? true : false;
	  }
	  if( false == isvalide)			// Erreur : chercher encore 
	  {
		m_fp.lastpos();
		cdata << "&lt;" ;				// conserver comme escape
		state= CX_TYPEUNDEF;			// continuer la lecture 
		break;            
	  }

	  break;							// tagname ok !

	}/* switch */

	if( CX_TYPEUNDEF==state ) continue;	  // Boucler


  // Casuel d'installation des noeuds dans l'arboresence
    switch( state)
    {                                     // analyse de balise d'ouverture
    case CX_PI_CMD:                 // processing instruction

      _eventdata(tagname, cdata);
      _eventdoc(state, tagname, NULL);
      _eventatt(state, balise);         // parsing des attributs 
      break; 

    case CX_ELEMENT:
	  /* Notes:
		- Les donnes cdata ne sont installes que sous un bloc (tagbloc!="")
		- Hors-bloc (tagbloc=="") les donnes cdata sont ignorees
		- Si dans un bloc, le tagname==tagbloc, cela signifie que le bloc 
		  prcdent n'a pas t referm. 
		  Dans ce cas, les donnes accumules sont galement ignores
	  */

	  if( tagbloc.empty())			// on se trouve hors-bloc ?
	  {
		// dtection automatique ou selon indications utilisateur
		if( m_usertag.empty() || !m_usertag.icompare(tagname) )		
		  tagbloc= tagname;			// signale un nouveau bloc
		else
		  break;					// balise de bloc non reconnue
	  }
	  else							// on se trouve intra-bloc ?
	  {
		if( !tagbloc.icompare (tagname))	// bloc prcdent non cltur ?
		{
		  _eventdoc(CX_ENDBLOCK, tagbloc, NULL);	// forcer la clture
		}
		else  // installer les donnes accumules sous le noeud prcdent
		{
		  _eventdata(tagname, cdata);  
		}
	  }

	  // installer un nouveau noeud 
      _eventdoc(state, tagname, NULL);
      _eventatt(state, balise);				// parsing des attributs 

	  // Faut-il ajouter un attribut id ?
	  if( m_withids)
	  {
		//  tous les noeuds ou  node racine seulement ?
		if( (2 & m_withids) || !tagname.icompare(tagbloc) )
		{
		  c_Str idvalue;
		  idvalue << "ID" << m_baseids++;	
		  _eventdoc(CX_ATTRIBUTE, "id", idvalue);
		}
	  }
      
	  break;

    case CX_ENDBLOCK:               // Balises de fermeture explicite

	  // Hors d'un bloc, les balises fermantes sont ignores
	  if( tagbloc.empty())	break;				  // (balise perdue ?)

	  _eventdata(tagname, cdata);				  // installer CX_TEXT
      _eventdoc(state, tagname, NULL);	  // installer

      if( !tagname.icompare(tagbloc))      // Fin de bloc-root ?
        tagbloc.erase();			// signale hors-bloc

      break;

    case CX_TAGEND:                 // balise de fermeture implicite

	  // Hors d'un bloc, les balises implicites sont ignores
	  if( !tagbloc.empty())	break;				

	  _eventdata(tagname, cdata);				  // installer CX_TEXT
      _eventdoc(CX_TAGEND, tagname, NULL);
      _eventatt(state, balise);         
      break;

    case CX_CDATA:                        // CDATA nodes 
    case CX_COMMENT:                      // COMMENT nodes 

      _eventdata(tagname, cdata);
      _eventdoc(state, tagname, balise);
      break; 

    default:  

      throw err( CX_ERR_ANY, "_xemparse: casuel de tag non prevu ! :))") ; 

    } /* switch(state) */

	// Nettoyer les donnes non retenues
	cdata.erase();

  }/* while (file) */


  // Vrification finale 
  // Notes:
  // Les donnes CX_TEXT accumules (par oubli de balise de fermeture) sont IGNOREES
  // Un warning est signal. 
  // 
  if( !tagbloc.empty()) 
  {
	_eventdoc(CX_ENDBLOCK, tagbloc, NULL);	// forcer la clture
    return err( CX_ERR_EOF, "Balise de fermeture de bloc manquante.");
  }
  else
	return err( CX_OK, "Ok.");

}

/*----------------------------------------------------------
  Ajouter les donnes accumules
  Les donnes sont celles qui prcdent le noeud 
  ----------------------------------------------------------*/

void c_Xem::_eventdata(c_Str& tagname, c_Str& cdata)
{
    if( cdata.empty()) return;            // Donnes accumulees ?

    cdata.strip();                        // purge des espaces
    if( NL==cdata[0]) cdata.ltrim();      // Pas de NL extraneous
    if( NL==cdata[-1]) cdata.chop();
    if( cdata.empty()) return;

	_eventdoc(CX_TEXT, "", cdata);	 // installer un CX_TEXT
    cdata.erase();           
}

/*----------------------------------------------------------
  Traitement des attb (CX_ELEMENT uniquement)
  ----------------------------------------------------------*/

void c_Xem::_eventatt(CX_TYPE context, c_Str& balise)
{
  c_Str attname= ""; 

  while(true)
  {
    // lire le nom de l'attribut (attention  ne pas renvoyer le PI "?>")
    attname= balise.tokennext(" \n=?", C_STRTOK);   
    if( attname.empty()) 
      break;

    attname.lower();      // toujours en minuscules

    // lire la valeur de l'attribut
    c_Str attdata= balise.tokbalance("\"'");   // lire le paramtre 
    attdata.trim();                

	// Ajuster la valeur de m_baseids si un id est dfinit
	if( m_withids && (attname== "id"))
	{
	  int idvalue= attdata.value(2);  // position 2 de "ID<value>"
	  if( idvalue >= m_baseids)
		m_baseids= idvalue+1;		  // premier libre
	}


    _eventdoc(CX_ATTRIBUTE, attname, attdata);  
  }
}


/*----------------------------------------------------------
  Installation des elements
  ----------------------------------------------------------*/
/*
  les noeuds sont toujous appended  c_Link::last()
*/

void c_Xem::_eventdoc(CX_TYPE context, lpstr psz1, lpstr psz2)
{
c_Xemnode* pnode= NULL;
c_Xemnode* parent= NULL;

  switch(context)
  {

  case CX_PI_CMD:               // psz1 = tagname / psz2= donnes dans balise
    // Installer *avant* le premier ELEMENT
    if( ! (pnode= new c_Xemnode()))
      throw err( CX_ERR_MEMORY, "memoire insuffisante");

    c_Link::insert(firstelement(), pnode);       // sur la racine     

    // todo - verifier duplicatas ??
    pnode->m_type    = context;     // CX_ELEMENT/CX_CDATA/CX_PICMD
    pnode->m_tagname = psz1;        // tag en clair
    pnode->m_cdata   = psz2;        // donnes lues
    pnode->m_lock= true;
    break;

  case CX_ATTRIBUTE:                // psz1= att-name psz2= att-data
    // Les attributs sont installs sur le dernier ELEMENT cr
	// Le test de prsence de l'attribut 'id' se fait sur tous les noeuds
    pnode= CAST_XEMNODE(c_Link::chain());
	if( !strcmp("id",psz1) && (-1 != pnode->m_attname.find(psz1)))          // nom de l'attribut
	  return;
    pnode->m_attname << psz1;          // nom de l'attribut
    pnode->m_attdata << psz2;          // valeur de l'attribut 
    break;

  case CX_ELEMENT:              // psz1= tagname, pz2= NULL
    // installer le noeud sur la RACINE !! 
    if( ! (pnode= new c_Xemnode()))
      throw CX_ERR_MEMORY;

    c_Link::add(NULL, pnode);       // sur la racine     
    pnode->m_type    = context;     // CX_ELEMENT/CX_CDATA/CX_COMMENT
    pnode->m_tagname = psz1;        // tag en clair
    pnode->m_lock    = false;
    break;

  case CX_ENDBLOCK:					// psz1 = tagname, pz2= NULL
     /* Balise fermante explicite: 
        Recherche le noeud racine unlocked le plus ancien
        Deplace les elements accumuls sous le noeud resultat
        Dfinit un lock
     */
    parent= CAST_XEMNODE(c_Link::last());        // recherche tag ouvrant
    while( parent && (parent->m_lock || parent->m_tagname!= psz1))  
        parent= CAST_XEMNODE(parent->prevsibling());

    if( !parent) 
      break;  // LEGAL: IGNORE SI PAS DE PARENT

    parent->m_lock= true;
    while( parent->nextsibling())      // deplace les noeuds sous le parent
      c_Link::movetochild(parent, parent->nextsibling()); 

    break;

// les noeuds CX_TEXT, COMMENT, CDATA et C_TAGEND 
// ne suivent pas le principe des balises ouvrantes/fermante et
// sont installs sous le prcdent 

  case CX_TEXT:						// psz1 = NULL / psz2= donnes accumulees
    // installer le noeud sous le prcdent
    if( ! (pnode= new c_Xemnode()))
      throw CX_ERR_MEMORY;
    c_Link::add(c_Link::last(), pnode);            
    pnode->m_type    = context;     // CX_ELEMENT/CX_CDATA/CX_PICMD
    pnode->m_tagname = "";          // pour dbug ou optionnel ? (unused)
    pnode->m_cdata   = psz2;        // donnes lues
    pnode->m_lock= true;
    break;

  case CX_TAGEND:                 // psz1 = tagname, pz2= NULL
  case CX_COMMENT:				  // psz1 = "" / psz2= commentaire 
  case CX_CDATA:				  // psz1 = "CDATA" / psz2= contenu de donnes
    // installer le noeud sous le prcdent
    if( ! (pnode= new c_Xemnode()))
      throw CX_ERR_MEMORY;
    c_Link::add(c_Link::last(), pnode);            

    pnode->m_type    = context;     // CX_ELEMENT/CX_CDATA/CX_COMMENT
    pnode->m_tagname = psz1;        // tag en clair
    pnode->m_cdata   = psz2;        // donnes lues
    pnode->m_lock= true;
    break;



  default: 
     throw err( CX_ERR_ANY, "_eventdoc:casuel incorrect :))");
  } 

  return;
}


/* eof */

