/*****************************************************
  file    : c_defs.h
  usage   : macros et decl de traitement de chaines
  date    : 1991 -- 
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  :
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/

#ifndef _C_DEFS_H_
#define _C_DEFS_H_

// Includes systmatiques
#include    <c_ascii.h>
#include    <c_char.h>

#include    <stdlib.h>

// TODO: ajouter une librairie GNU
#ifdef WIN32
#   pragma warning(disable:4305)    //'default argument' : truncation from 'const int' to 'bool
#   pragma warning(disable:4800)    //'default argument' : truncation from 'const int' to 'bool
#endif 





// Gestion des DLL
// __declspec works in BC++ 5 and later, as well as VC++ and gcc
// -------------------------------------------------------------
#    define C_DLLEXPORT            __declspec( dllexport )
#    define C_DLLEXPORT_DATA(type) __declspec( dllexport ) type



/*[1]	Manifestes constantes (1)
/*-------------------------------------------------*/

#ifndef NULL
#ifdef __cplusplus
#define NULL    0
#else
#define NULL    ((void *)0)
#endif
#endif

typedef unsigned long	    ulong;
typedef unsigned int      uint;
typedef unsigned short    ushort;

// Utilise pour les formatages de c_File>> hexa_8;
typedef unsigned char     hexa_8;
typedef unsigned short    hexa_16;
typedef unsigned long	    hexa_32;


#define elif		else if
#define defaut	default		    /* compiler ne rend pas d'erreur! */
#define repeat	while(1)	    /* pour for(EVER) */
#define EVER	  ;;	          /* pour for(EVER) */
#define NIL		  NULL		      /* Equivalenet de NULL */


/*[-]	Gestion de mmoire
/*************************************/
#define c_malloc(cnt)            ((cnt)?malloc(cnt):NULL)
#define c_memmove(ad1,ad2,cnt)   (((cnt)&&(ad1!=ad2))?memmove(ad1,ad2,cnt):ad1)
#define c_memcpy(ad1,ad2,cnt)    (((cnt)&&(ad1!=ad2))?memmove(ad1,ad2,cnt):ad1)
#define c_memset(ad,ch,cnt)      ((cnt)?memset(ad,ch,cnt):ad)
#define c_strdup(p)              (((p)&&*(p))? strdup(p):NULL)

/*[-]	Redfines de fonctions
/*************************************/
// __BORLANDC__	  version Borland
// _MSC_VER       version Microsoft

#ifdef _MSC_VER
#define	strlwr	_strlwr
#define	strupr 	_strupr
#endif


/*[-]	Intervalles et majeur/mineur numeriques
/*************************************/
#define c_interval(n,l,h)   (((n)>=(l) && (n)<=(h) )? true: false)
#define c_max(a,b)          (((a) > (b)) ? (a) : (b))
#define c_min(a,b)          (((a) < (b)) ? (a) : (b))
#define c_major( n1, n2)    c_max(n1,n2)
#define c_minor( n1, n2)    c_min(n1,n2)


/*[-]	Manipulations sur b-maps
/*************************************/

#define c_valbit(b)	      ( 1L<<(unsigned)(b))
#define c_valbit_l(b)	    ((ulong) (1L<<(unsigned)(b) ) )
#define c_valbit_w(b)	    ((ushort)( 1<<(unsigned)(b) ) )
#define c_valbit_c(b)	    ((uchar) ( 1<<(unsigned)(b) ) )

#define c_cplbit(b)	      (-1L ^ c_valbit(b)   )
#define c_cplbit_l(b)	    (-1  ^ c_valbit_l(b) )
#define c_cplbit_w(b)	    (-1  ^ c_valbit_w(b) )
#define c_cplbit_c(b)	    (0xFF^ c_valbit_c(b) )

#define c_isbit( n,b)	    ( ((n)& c_valbit((b)))? true:false  )
#define c_isbit_l( n,b)	  ( ((n)& c_valbit_l((b)))? true:false)
#define c_isbit_w( n,b)	  ( ((n)& c_valbit_w((b)))? true:false)
#define c_isbit_c( n,b)	  ( ((n)& c_valbit_c((b)))? true:false)



/*[-]	Hard-Code Conversion
/*-------------------------------------------------*/
/** (short) Binaire coded Decimal: un digit binaire per quartet
*/
#define DEC_BCD(x)	    ( ((((short)x)/10)<<4) + (((short)x)%10) )
#define BCD_DEC(x)	    ( (((short)x)&0xF)+10*(((short)x)>>4) )

/* (short)Binaire coded Ascii-digit: un DIGIT ascii per octet
*/
#define DEC_BCA(x)	    (0x3030| (((((short)x)/10)<<8)+(((short)x)%10)))
#define BCA_DEC(x)	    ( ((x)&0xF)+10*(((x)>>8)&0xF) )

#define LSQ(b)		      (uchar)(0xF & (b))
#define MSQ(b)		      LSQ(((uchar)(b))>>4)
#define MK_BYTE(lo,hi)	((LSQ(hi)<<4) | LSQ(lo))
#define MK_SWAQ(b)	    (uchar) MK_BYTE( MSQ(b), LSQ(b))

#define LSB(b)		      (uchar)(0xFF & (b))
#define MSB(b)		      LSB(((ushort)(b))>>8)
#define MK_WORD(lo,hi)	((LSB(hi)<<8) | LSB(lo))
#define MK_SWAB(b)	    (ushort) MK_WORD( MSB(b), LSB(b))

#define LSW(b)		      (0xFFFFL &(long)(b))
#define MSW(b)		      LSW(((ulong)(b))>>16L)
#define MK_LONG(lo,hi)	(ulong) ((LSW(hi)<<16L) | LSW(lo))
#define MK_SWAW(b)	    (ulong) MK_LONG( MSW(b), LSW(b))



/*[-1]	 DATE_t avec conversion
/*-------------------------------------------------*/
/*  Bit structure:
/*
/*  b0-b4  <32	day: [0..30]
/*  b5-b8  <16	mon: [0..11]
/*  b9-b15 <128 year:[0..127]	annees depuis tunable YDATE
/*-------------------------------------------------*/
typedef ushort DTM_t;

#define YDATE 1992
#define DTM_day(dtm)	(1+(((DTM_t)(dtm))&0x1F))	/* b0-b4  <32	*/
#define DTM_mon(dtm)	(1+((((DTM_t)(dtm))>>5)&0xF))	/* b5-b8  <16	*/
#define DTM_year(dtm)	(((DTM_t)(dtm))>>9)		/* b9-b15 <128	*/

#define DTM_YEAR(dtm)	( (YDATE+DTM_year(dtm)))  /* format YDATE+year */
#define DTM_YY(dtm)	  ( DTM_YEAR(dtm)%100)	    /* format YY	*/


/*[-] ukey_t & MK_UKEY(k,i)
/*-------------------------------------------------*/
/*  Macro d'usage general pour determiner le key d'un c_umalloc()
/*  Typiquement: hi=key correspond a la classe de l'objet
/*		 lo=id	correspond au n de serie de l'objet de cette classe
/*-------------------------------------------------*/
typedef unsigned long ukey_t;
#define MK_UKEY(id,class)    (ukey_t)(MK_LONG(id,class))

/* TAGS declaration*/
#define GETTAG_l(t,p)	 memcpy(&t,(ulong*)(p),sizeof(ulong))
#define GETTAG_c(t,p)	 memcpy(&t,(char*)(p),sizeof(ulong))



#endif  /* _C_DEFS_H_*/
