/*****************************************************
  fichier : c_file.cpp
  usage   : classe extension de FILE*
  date    : 1991--
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  : ideas from wxWindows (ungetch)
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/


#ifndef _C_FILE_H_
#define _C_FILE_H_

#include <c_defs.h>
#include <c_string.h>

#include <stdio.h>
#include <stdarg.h>

/*-------------------------------------*/
/*  Defined  c_filefn.cpp
/*-------------------------------------*/
typedef char PATH_t[128];
C_DLLEXPORT lpstr c_fcheck    ( c_Str& p);
C_DLLEXPORT lpstr c_fcheck    ( lpstr p);


// Bitmap d'options pour c_freadpath
#define	C_DIR	4	
#define	C_NAME	2
#define	C_EXT	1
C_DLLEXPORT lpstr   c_splitpath ( lpstr p, int options);// -dir +name +ext
C_DLLEXPORT lpstr   c_makepath  ( lpstr p, lpstr b, lpstr e);

/*  unix compatibles ?
/*	c_fdependency    ( lpstr f1, lpstr f2)	    =0 si f1>f2,  -1 sinon
/*	c_fexist	    ( lpstr )		    =0 si existe  -1 sinon
/*  long   c_fsize	    ( lpstr )		    >0 si size	  -1 sinon
/*	c_fcopy	    ( lpstr in,lpstr out,int) =0 si ok	  -1 sinon
*/
C_DLLEXPORT bool    c_fdependency ( lpstr f1, lpstr f2);
C_DLLEXPORT bool    c_fexists     ( lpstr f1);
C_DLLEXPORT size_t  c_fsize       ( lpstr f1);
C_DLLEXPORT bool    c_fcopy       ( lpstr inp, lpstr out, bool mode);


/*-------------------------------------*/
/*  Defined  c_file.cpp
/*-------------------------------------*/
class C_DLLEXPORT c_File
{

private:
    FILE*   m_fp;       // IO stream or NULL if not opened
    c_Str   m_name;     // the name of the file (for diagnostic messages)
    fpos_t  m_pos;      // savepos/lastpos()


    // Empcher la dfinition du cpy-ctor par l'utilisateur
    c_File  (const c_File&);
    c_File& operator=(const c_File&);

    // Ungetch managers
    void    _reset()            
    { 
      if(m_wback) free(m_wback); 
      m_wback= NULL; 
      m_wbackcur= m_wbacksize= 0;
    }
    lpstr   m_wback;
    size_t  m_wbackcur;
    size_t  m_wbacksize;
    lpstr   _allocback  (size_t needed_size);
    int     _getback    ();   /* retourne -1 si vide */

public: 
    ~c_File()         { close(); if (m_wback) free(m_wback); }
    c_File()          { m_fp= NULL; m_wback= NULL; _reset(); }
    c_File(FILE *fp)  { m_fp= fp; m_wback= NULL; _reset(); }
    c_File(lpcstr filename, char mode = 'r') 
    { m_fp= NULL; m_wback= NULL; _reset(); open(filename,mode);}

    FILE* fp()      const { return m_fp; };
    bool eof()            { return (m_fp)? feof(m_fp): true; } 
    bool flush()          { _reset(); return (m_fp)? (EOF!=fflush(m_fp)) : false; }
    bool isopen()  const  { return m_fp != NULL; };
    void attach(FILE*fp)  { if( m_fp) close(); m_fp= fp; }
    void detach()         { m_fp= NULL; }

    /* Noter que la syntaxe est un CHAR: toujours ouvert en binaire */
    bool open   (lpcstr filename, char mode = 'r'); 
    bool close  ()    { m_name.erase(); return m_fp ? (!fclose(m_fp)) : false; }
    size_t length() const;

public: // deplacement (avec _reset())

    bool seek   (long ofs, long mode = SEEK_SET) 
    { _reset(); return (m_fp && !fseek(m_fp,ofs,mode)) ? true: false; }
    bool seekend(long ofs = 0)      {return seek(ofs, SEEK_END); }
    bool seekcur(long ofs = 0)      {return seek(ofs, SEEK_CUR); }
    bool seekset(long ofs = 0)      {return seek(ofs, SEEK_SET); }
    bool rewind  ()                 {return seek(0L); } 

    size_t  getpos() const            /* utilise fgetpos!!! */
    { 
      if (!m_fp) return -1; 
      return ftell(m_fp) - (m_wbacksize-m_wbackcur); 
    }
    size_t  setpos (size_t pos)     { _reset(); return seek(pos,SEEK_SET); }

    void    savepos ()              { fgetpos(m_fp, &m_pos); }
    void    lastpos ()              
    { 
      _reset(); 
      fpos_t a_pos;
      fgetpos(m_fp, &a_pos);
      if( a_pos > m_pos)    // MS-BUG? Eviter resync avec 0 offset
      {
        fsetpos( m_fp, &m_pos); 
      }
    }
    size_t  diffpos () const        
    { 
      fpos_t a_pos;
      fgetpos(m_fp, &a_pos);
      return (size_t)( a_pos-m_pos); 
    }

    // Deplacement avec getch()) 
    bool     seek_not_of   ( lpcstr szchars);	      // ignore caractres
    bool     seek_no_blanks();                          // ignore espace ET NL!!!!

private: 
    bool    _seekstr    ( lpcstr sequence, c_Str* echo);
public:                 
    bool     seekstr	  ( lpcstr sequence)
                        { return _seekstr(sequence, NULL); }
    bool     seekstr	  ( lpcstr sequence, c_Str& echo)
                        { return _seekstr(sequence, (c_Str*) &echo); }

    bool     seekstr	  ( lpcstr sequence, c_File& echo);

    bool     seekeol    ();              // aprs nl
    bool     seekeol    ( c_Str& str) { return seekstr("\n", str); }
    bool     seekline   ( int lines);	   // rewind + positionne sur n ligne


    bool     seektag    ( lpcstr tag, c_Str* echo= NULL);
    bool     seektag    ( lpcstr tag, c_Str& echo) { return seektag(tag,&echo); }


public: // Lecture avec gestion de buffer dynamique pour ungetch
    c_File&  operator>>(c_Str& str){ seekeol(str); return (*this); }    
    c_File&  operator>>(uchar ch)  { ch= getch(); return (*this); }
    c_File&  operator>>( char ch)  { ch= getch(); return (*this); }
    c_File&  operator>>(int val)   { val = getch(); return (*this); }
    c_File&  operator>>(long val)  { val= getnumber(); return (*this); }

    int      read       (lpstr  buffer, int    count);     
    int      read       (c_Str& buffer, int    count)
    {
      size_t nret;
      buffer.reserve(count);
      nret = read( (lpstr) buffer, count);
      buffer.sync();      // resynchronise m_end()
      return nret;
    }

    uchar    peek       ();
    int      peek       (lpstr  buf, size_t count);
    int      peekline   (c_Str& buf);   

    uchar    getch      ();	     /* CRLF2LF conversion */
    bool     ungetch    (char  ch); 
    int      ungetch    (lpcstr buf, size_t bufsize);
    int      ungetch    (lpcstr buf) { return ungetch(buf, strlen(buf)); }
    int      ungetch    (c_Str& buf) { return ungetch(buf, buf.size()); }

    bool    getline   (c_Str& buf)  { return seekeol(buf); }
    bool    getlinenl (c_Str& buf)  
    { 
      if( false==seekeol(buf)) return false;
      buf<< NL; 
      return true;
    }
    bool    getfile   (c_Str& buf); 
    long    getnumber (int count=0);   // Syntaxe hexa. dfaut longueur indtermine.

//  int     GetFormatsRTF( c_Str& szBuffer);
//  int	    GetParagraphe( lpstr szBuffer, int nSize);

public: // Ecriture: cf wxInputStream
/* Interface C standard 
/*  Pas de curseur dans ce contexte
*/
    c_File&  operator<<(uchar ch)   { putch(ch); return (*this); }    
    c_File&  operator<<( char ch)   { putch(ch); return (*this); }    
    c_File&  operator<<(lpcstr str) { puts(str); return (*this); }    
    c_File&  operator<<(c_Str& str) { write(str); return (*this); }    
    c_File&  operator<<(uint val)   { puthexa(val,4); return (*this); }    
    c_File&  operator<<(ulong val)  { puthexa(val,8); return (*this); }    

    void    putch   (char c, int n=1)       { if(m_fp) while( n--) ::fputc( c, fp()); }
    void    puts    (lpcstr buf)            { while( *buf) putch(*buf++); }
    void    putsnl  (lpcstr buf)            { if(buf) puts(buf); putch(NL); }
               
    size_t  write   (lpcstr buf,int size)   { return (m_fp)? (::fwrite(buf,1,size, m_fp)) : 0; }
    void    writenl (lpcstr buf,int size)   { write(buf,size); putch(NL); }
    size_t  write   (c_Str& buf)            { return write((lpcstr)buf, buf.size()); }

    // E/S d'entiers
    bool putint(long prmt)     // Format de %0.8X pour base 16
            { 
                fprintf( fp(), "%d", prmt);
                return true;  // TODO
            }
    bool puthexa(long prmt, int pad=0)     // Format de %0.8X pour base 16
            { 
                switch(pad)
                {
                case 4: fprintf( fp(), "0%0.4X", prmt); break;
                case 8: fprintf( fp(), "0%0.8X", prmt); break;
                default:fprintf( fp(), "0%X", prmt); break;
                }
                return true;  // TODO
            }

    // Format 
    int     printf(lpcstr format, ...)          
            { 
            va_list     prmt;
            int         retval;
                va_start( prmt, format);
                retval= vfprintf( fp(), format, prmt);
                va_end(prmt);
                return retval;
            }
    int     printfnl(lpcstr format, ...)     // Rajoute NL!
            { 
            va_list     prmt;
            int         retval;
                va_start( prmt, format);
                retval= vfprintf( fp(), format, prmt);
                va_end(prmt);
                putch(NL);
                return retval;
            }


};


#endif // _C_FILE_H_