/*****************************************************
  file    : c_Link.h
  purpose : liens d'arborescence
  date    : 1995 --
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  :
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/

#ifndef _c_LINK_H_
#define _c_LINK_H_

/*----------------------------------------------------------

  Manual:

*/	



/*----------------------------------------------------------*/

#include <stdlib.h>

class  c_Node;
class  c_Link;

/*----------------------------------------------------------*/
/*  c_Node
/*  Abstract classe: tout node ajout  c_Link doit driver d'un c_Node
/*----------------------------------------------------------*/

class c_Node
{
friend class  c_Link;       // Accessible  c_Link (pas  l'objet!)

    size_t  m_ukey;         // unique, si !0
    bool    m_match;        // si SELECT

    c_Node  *m_next;	      // (chainage historique)
    c_Node  *m_prev;	      //
                            // arborescence
    c_Node  *m_parent;	    // Node primaire
    c_Node  *m_child;       // Premier node secondaire (liste NULL)
    c_Node  *m_prevsib;	    // Node frre anterieur de mme Parent (liste NULL)
    c_Node  *m_nextsib;	    // Node frre suivant de mme Parent (liste NULL)

public:
    c_Node()  { m_next= m_prev= m_parent= m_child= m_prevsib= m_nextsib= NULL; };
    virtual ~c_Node() {};
    virtual bool isvalid() =0; // { return false; }

    size_t  ukey ()                   { return m_ukey; }
    bool    match(bool b=true)        { return (b)? (m_match=b) : m_match; }

    // Reprages
    c_Node  *next () { return m_next; }
    c_Node  *prev () { return m_prev; }
    bool    isbeg () { return (NULL==m_prev) ? true : false ;}
    bool    isend () { return (NULL==m_next) ? true : false ;}

    // Deplacement sur arbre
    c_Node  *parent       () { return m_parent; }
    c_Node  *child        () { return m_child; }
    c_Node  *nextsibling  () { return m_nextsib; }
    c_Node  *prevsibling  () { return m_prev; }
    bool    isbegsibling () { return (NULL==m_prevsib) ? true : false ;}
    bool    isendsibling  () { return (NULL==m_nextsib) ? true : false ;}
    size_t  level () // distance de root 
    {
      c_Node* pnode= this;
      size_t  count;
      for( count= 0; pnode->m_parent; count++)
        pnode= pnode->m_parent;
      return count;
    }
};


/*----------------------------------------------------------*/
/*  Linkage
/*----------------------------------------------------------*/

class c_Link : public c_Node  // !Permet de grer des c_Link de c_Link(s)!
{

  // Structure de base     ---------------------------

private:
    size_t     m_count;	      // Nombre de nodes
    size_t     m_keygen;	    // gnrateur de keys
    c_Node    *m_root;	      // Node racine (et dbut des m_Liens)
    c_Node    *m_chain;	      // repere historique, dernier node install

public:
    c_Link(size_t base_key= 0)  
    { 
      m_keygen= (base_key) ? base_key: 0x10000;   // JAMAIS =0
      m_count= 0; 
      m_chain= m_root= NULL; 
    }
    virtual ~c_Link ()    {   clear(); }
    virtual bool isvalid(){  return true; }

    void  clear();


  // Structure d'iteration ---------------------------

private:
    c_Node* _init      (c_Node* obj, size_t ukey);        // usage interne : vrai si 1e cration
    void    _erase    (c_Node* obj);      // usage interne 

    int     _sort      (c_Node* p1, c_Node* p2);  // helper de sort()
    bool    _renum     (c_Node* pnode); // Helpers de parse()
public:

    // Accessors
    c_Node* root    ()const  { return m_root; }
    c_Node* chain   ()const  { return m_chain; }       // Dans l'historique
    c_Node* last    (c_Node* parent= NULL)const;  // TreeCtxt
    size_t  keygen  (size_t val=0) { return (val) ? m_keygen=val: m_keygen; }

    size_t  count (c_Node* obj= NULL)  const;    // Nombre de descendants ou m_count
    c_Node* root  (size_t ukey)const { return m_root; }
    c_Node* find  (size_t ukey)const;
    bool    exists(size_t ukey)const { return find(ukey) ? true : false; }

    // Insertions 
    bool add          (c_Node* parent, c_Node* obj, size_t ukey=0);     
    c_Link& operator<<(c_Node* obj) { add(NULL, obj, 0); return *this; }
    bool insert       (c_Node* pto, c_Node* obj, size_t ukey=0);     // Inserer devant here
    void moveto       (c_Node* pto, c_Node* obj);  // sibling
    void movetochild  (c_Node* pto, c_Node* obj);  // ajoute aux childs du node



    // Effacement
    bool   erase  (c_Node* obj= NULL);  // Efface node+descendants+Sync()
    bool   erase  (size_t ukey)   {return erase(find(ukey)); }

    // Parsing - TreeContext 
    virtual bool  eventparse(c_Node* obj) { return true; } // default
    bool parse    (c_Node* obj= NULL, bool(*func)(c_Node* p)= NULL);  

    // Sorting (TreeContext: selon ukey par dfaut) 
    virtual int   eventsort(const c_Node* p1, const c_Node* p2); 
    bool sort     (c_Node* obj= NULL, int(*func)(const c_Node* p1, const c_Node* p2)= NULL);  

    // Matching TreeContext
    void   match   (c_Node* obj= NULL, bool state= true); 
    void   match   (size_t ukey, bool state= true) { match(find(ukey), state); }
    void   matchall(c_Node* obj= NULL, bool state= true); 
    void   matchall(size_t ukey, bool state= true) { matchall(find(ukey), state); }

    // Renumrotation TreeContext (defaut m_keygen)
    bool   renum   (c_Node* obj, size_t newkey); 
    bool   renum   (size_t newkey) { return renum( NULL, newkey); }

};


/******************************************************/
#endif // _c_LINK_H_



