/*****************************************************
  file    : c_Str.h
  purpose : chaine dynamique avec extensions std::string/stream
  date    : 1991 --
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  :
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/


#ifndef _C_STR_H_
#define _C_STR_H_

#include <string.h>           // strdup
#include <c_defs.h>           // c_Ascii + c_Char

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------


#define C_STR_NPOS    0x7FFFFFFF    // basic_string::npos 
#define C_STR_MAXSIZE 0x7FFFFFFE    // basic_string::maxsize (32K

// Paramtres de c_Str::token(mode)
typedef enum
{ 
  C_KEEPTOK=0   /* conserver le contexte, "tok1::tok2:" = 3 tokens*/
, C_DFTTOK      /* defaut mode, "tok1::tok2:" = 3 tokens*/
, C_WDELIM      /* append delim "tok1:", ":", tok2:" = 3 tokens*/
, C_STRTOK      /* standard     "tok1", tok2:" = 2 tokens*/
} C_TOKMODE;

// Bitmap de c_Str::trim(mode)      (tester avec c_isbit!!!)
enum
{ 
  C_TRIMRIGHT=1    /* defaut mode, "tok1::tok2:" = 3 tokens*/
, C_TRIMLEFT =2    /* append delim "tok1:", ":", tok2:" = 3 tokens*/
};


// Bitmap de c_Str::htmencode(mode) (tester avec c_isbit!!!)
enum
{
  C_HTMLPACK  =1   // packing des blancs et NL rpts
, C_HTMLALIAS =2   // utilise aliases d'entits HTML
, C_HTMLSTRICT=4   // utilise entits numriques strictes
}; 



/*
  Classe de chaine sans contrle d'erreur et accessible en criture
  Rapide. Usage interne.
*/

class C_DLLEXPORT c_Str
{
  lpstr   m_data;	      // donnes
  lpstr   m_end;	      // pointeur d'criture
  int 	  m_mem;	      // allou arrondis au C_CHARBLOCK suprieur

  char    m_tokch;      // replace char
  int     m_tokpos;     // pointeur chaine retournee C_STR_NPOS si fin de recherche
  lpstr   m_tokstr;      // chaine de tokening 
  C_TOKMODE m_tokmode;  // C_DFTTOK ou C_DELTOK 
  bool    m_tokspc;     // true si contient uniquement " \t\n"
  lpstr   _token(lpcstr szDelim, C_TOKMODE mode);  // traitement proprement dit
  void    _tokraz() { m_tokch=  m_tokpos= 0; m_tokstr= NULL; }


  bool    _memalloc(int  bytes);
  bool    _init( lpcstr sz, int pos, int count);
  void    _clear();  

public: // ctor

  ~c_Str()		    { _clear();  }
  c_Str()		                                      { _init(NULL,0,0);  }     
  c_Str( lpcstr sz, int  nLen = C_STR_MAXSIZE) { _init(sz,0,nLen);    }       
  c_Str( lpcstr sz, int pos, int nLen = C_STR_MAXSIZE)  { _init(sz,pos,nLen);  }       
  c_Str( char  ch, int  count =1);
  c_Str( long val, int base=10);

  c_Str& operator= (const c_Str& sz) { return operator= (sz.m_data); }
  c_Str& operator= (lpcstr sz);
  c_Str& operator= (char ch);
  c_Str& operator= (int n);
  c_Str& operator= (long n);

public: // std::basic_string semantique
  int maxsize()        {   return C_STR_MAXSIZE; }
  lpcstr c_str()          {   return m_data; }

  int capacity()	      {   return m_mem; }
  lpstr  data()	          {   return m_data; }
  lpstr  end()	          {   return m_end; }
  int size()	          {   return (m_data) ? (m_end-m_data): 0; }
  int length()	        {   return (m_data) ? (m_end-m_data): 0; }
  bool   empty()          {   return size() ? false : true; }

  c_Str& append (lpcstr sz, int pos, int count= C_STR_MAXSIZE);
  c_Str& append (lpcstr sz, int count= C_STR_MAXSIZE)
  { return append (sz, 0, count); }
  c_Str& append (int count, char ch);

  c_Str& erase  (int pos=0, int count= C_STR_MAXSIZE);

  void   reserve(int  count);
  void   resize (int  count, char ch=0);


  c_Str& operator<< (char ch);        // append
  c_Str& operator<< (int  ch);        // append decimal
  c_Str& operator<< (long ch);        // append decimal
  c_Str& operator<< (lpcstr sz);	            
  c_Str& operator<< (c_Str& sz);	            

  c_Str& operator+  (char ch)	{ return operator<<(ch); };       
  c_Str& operator+  (int  ch)	{ return operator<<(ch); };       
  c_Str& operator+  (long ch)	{ return operator<<(ch); };       
  c_Str& operator+  (lpcstr sz)	{ return operator<<(sz); };       	            
  c_Str& operator+  (c_Str& sz)	{ return operator<<(sz); };       	            

  char operator[] (int pos);	    // smart [-1] equivaut  last()
  char charat(int pos);	        // smart [-1] equivaut  last()

  c_Str& insert(int pos, long val, int base=10)       
  { return insert( pos, c_Str(val, base)); }
  c_Str& insert(int pos, int count, char ch);       // Insert
  c_Str& insert(int pos, lpcstr sz, int pos2, int  len2);  // Insert
  c_Str& insert(int pos, lpcstr sz, int  len2 = C_STR_MAXSIZE)
  { return insert(pos, sz, 0, len2); }
  
  c_Str& replace(int pos, int count, int val, int base=10)
  { return replace( pos, count, c_Str((long)val, base)); }
  c_Str& replace(int pos, int count, long val, int base=10)
  { return replace( pos, count, c_Str(val, base)); }
  c_Str& replace(int pos, int count, int count2, char ch)
  { return replace( pos, count, c_Str(ch, count2)); }
  c_Str& replace(int pos, int count1, lpcstr sz, int pos2, int  count2);
  c_Str& replace(int pos, int count1, lpcstr sz, int  count2 = C_STR_MAXSIZE)
  { return replace(pos, count1, sz, 0, count2); }


  c_Str  substr (int pos, int count= C_STR_MAXSIZE) const;  // return object

  int   compare( int pos, int count1, lpcstr sz, int pos2=0, int  count2 = C_STR_MAXSIZE) const;
  int   compare( lpcstr sz, int pos2=0, int  count2 = C_STR_MAXSIZE)
  { return compare( 0, C_STR_MAXSIZE, sz, pos2, count2); } // non standard :))
  
  bool operator==(lpcstr sz) const
  { return (0==compare( 0, C_STR_MAXSIZE, sz, 0, C_STR_MAXSIZE)) ? true : false; }
  bool operator!=(lpcstr sz) const { return !operator==(sz); }

  bool operator<(lpcstr sz) const
  { return (compare( 0, C_STR_MAXSIZE, sz, 0, C_STR_MAXSIZE)<0) ? true : false; }
  bool operator>(lpcstr sz) const
  { return (compare( 0, C_STR_MAXSIZE, sz, 0, C_STR_MAXSIZE)<0) ? true : false; }
  bool operator<=(lpcstr sz) const
  { return (compare( 0, C_STR_MAXSIZE, sz, 0, C_STR_MAXSIZE)<=0) ? true : false; }
  bool operator>=(lpcstr sz) const
  { return (compare( 0, C_STR_MAXSIZE, sz, 0, C_STR_MAXSIZE)<=0) ? true : false; }
  
  
  int   icompare( int pos, int count1, lpcstr sz, int pos2=0, int  count2 = C_STR_MAXSIZE) const;
  int   icompare( lpcstr sz, int pos2=0, int  count2 = C_STR_MAXSIZE) const
  { return icompare( 0, C_STR_MAXSIZE, sz, pos2, count2); }


  int ifind    (lpcstr sz, int pos=0) const;   // ignorecase find 

  int find  (char ch, int pos=0) const;
  int find  (lpstr sz, int pos=0) const;
  int rfind (char ch, int pos=C_STR_MAXSIZE) const;
  int rfind (lpstr sz, int pos=C_STR_MAXSIZE) const;

  int find_first_of(char ch, int pos=0) const;
  int find_first_of(lpcstr sz, int pos=0, int count= C_STR_MAXSIZE) const; 
  int find_last_of(char ch, int pos=0) const;
  int find_last_of(lpcstr sz, int pos=0, int count= C_STR_MAXSIZE) const;

  int find_first_not_of(char ch, int pos=0) const;
  int find_first_not_of(lpcstr sz, int pos= 0, int count= C_STR_MAXSIZE) const;
  int find_last_not_of(char ch, int pos=0) const;
  int find_last_not_of(lpcstr sz, int pos=0, int count= C_STR_MAXSIZE) const;

public: // Extensions

  operator lpstr()	      {   return m_data; }
  operator lpcstr()       {   return m_data; }
  long     value            ( int pos=0, int len=C_STR_MAXSIZE);

  void  attach(lpstr sz);   // manipulation de chaine (non libre)

  void   sync()           {   if( m_data) m_end= c_adr_nul(m_data);  }

  char   last()	          {   return (c_isvalid(m_data)) ? m_end[-1] : 0; }
  lpstr  last(int pos) 
  { 
    if( pos > (int)(m_end-m_data)) return m_data;
    return (c_isvalid(m_data)) ? m_end-pos : NULL;
  }
  lpstr  dup()	          {   return (m_data) ? c_strdup(m_data) : NULL; }
  char   pop();	                // Supprime et retourne dernier char (-1 si 0 ou NULL)
  lpstr  pop( int count);	  // Supprime et retourne dernier char (-1 si 0 ou NULL)

  int    trim     ( int mode= C_TRIMRIGHT|C_TRIMLEFT);   // tous les blancs < SPC
  int    chop     ( int mode= C_TRIMRIGHT) { return trim (mode); }     // tous les blancs < SPC
  int    ltrim    ( int mode= C_TRIMLEFT)  { return trim (mode); }     // tous les blancs < SPC
  int    trim     ( lpcstr szDelims, int mode= C_TRIMRIGHT|C_TRIMLEFT);

  c_Str  extract  (int pos, int count= C_STR_MAXSIZE);  // return object
  c_Str  shift    (int count= C_STR_MAXSIZE) {return extract(0,count); }

  lpstr  tokenstart (lpcstr szDelim, C_TOKMODE mode= C_STRTOK);
  lpstr  tokennext  (lpcstr szDelim= NULL, C_TOKMODE mode= C_STRTOK);
  int    toksync    ();          // restitue la chaine, retourne pos de contexte
  int    toksync    (int pos);   // restitue la chaine, positionne pos de contexte 
  char   toksep     ()  { return m_tokch; }   // dlimiteur retenu
  lpstr  tokbalance (lpcstr szSep); // sans position: contexte de tokening

  int    balance    (lpcstr szSep, int pos=0);    // balance l'un des chars

  bool   printf   (int maxlen, lpcstr format, ...); 
  bool   appendf  (int maxlen, lpcstr format, ...); 

  int findtag  (int pos, int len, lpcstr sz) const;   
  int findtag  (int pos, lpcstr sz) const { return findtag (pos, C_STR_MAXSIZE, sz);}
  int findtag  (lpcstr sz) const { return findtag (0, C_STR_MAXSIZE, sz);}

  void  upper() { if( c_isvalid(m_data)) strupr(m_data); }
  void  lower() { if( c_isvalid(m_data)) strlwr(m_data); }


public: // specifics

  void  strip     (); // Elimine espaces doubles. Prserve NL

  void  addslashes(); // escaping de quotes 

  void  htmldecode();            // conversion HTML - ISO-8859-1
  void  htmlencode(int mode);    // conversion ISO-8859-1 - HTML 

};
 

#endif // _C_STR_H_
