/*****************************************************
  file    : c_Vector.cpp
  usage   : vecteurs de bits, chaines et objets
  date    : 2001
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  :
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/

#ifndef _C_VECTOR_H
#define _C_VECTOR_H

#include <c_defs.h>
#include <c_string.h>

/*----------------------------------------------------------
  c_Vector
/*----------------------------------------------------------*/	

#define C_VITEM(u) (long)(u)  // [inp] cast
#define C_VPTR     (void*)    // [out] cast

class c_Vector
{
friend class c_Mappe;
friend class c_Strarray;
friend class c_Voidarray;

  ulong  *m_data;     // pointer to data
  int     m_size;     // mmoire alloue (multiple de (sizeof(long))
  int     m_count;    // nombre d'lments logiques


public:
  c_Vector ()  { m_size=m_count=0; m_data=NULL; }
  virtual ~c_Vector () { if( m_data) free(m_data); }   // appeler le drived

  int    count() const   { return m_count;      }
  bool   reserve(int m_elem);
  void   add     ( ulong item);
  void   insert  ( int pos, ulong item);
  void   erase   ( int pos=0, int len=0);  // 0== ALL
  void   pack    ();    // shrink

  int    find    ( ulong  item) const;     // -1 si pas trouv
  bool   exists  ( ulong  item) const { return (-1 == find(item)) ? false : true; }
  ulong  item    ( int pos) const;
  ulong  last    ( int pos=0) const { return item(-1); }

  c_Vector& operator<<(ulong item)  { add(item); return *this; }
  ulong     operator[](int pos) const { return item(pos); }
};



/*----------------------------------------------------------
  c_Mappe
/*----------------------------------------------------------*/
/*
  Mappe binaire utilisee pour SELECT("critres")
  ATTENTION RAISONNE PAR MULTIPLES DE 2 et sur la limite de 32bits
*/

class c_Mappe: public c_Vector
{
  uchar   m_ubits;      // nombre de bits par unit
  int     m_elem;      // nombre d'units binaires
  ulong   m_mask;      // masque d'operators

public:
  c_Mappe() {define(1); }     // 1 bit = 1 element par defaut
  void define(int units);

  int  count ();              // nombre d'units non nulles

  bool setmap(int pos, int val);       // positionne l'unit
  bool resmap(int pos);                // annule l'unit ou false si outbound
  int  getmap(int pos)const;                // valeur ou -1 si outbound
  int  operator[](int pos) { return getmap(pos); }

  c_Mappe& operator|=(const c_Mappe& v);
  c_Mappe& operator&=(const c_Mappe& v);
  c_Mappe& operator^=(const c_Mappe& v);

  c_Mappe& operator<<(const int val);
};



/*  Array de c_Str  (Tout inline)
    Attention se sont des POINTEURS
*/
class c_Strarray: public c_Vector
{
public:
  c_Strarray ()  { m_size=m_count=0; m_data=NULL; }
  virtual ~c_Strarray () { erase(); }
  void   erase(int pos=0, int len=0) 
  { 
    if( !m_count) return;
    if( pos<0) pos+= m_count;
    if( pos > m_count) return;
    if( !len) len= m_count;
    if( pos+len > m_count) len= m_count-pos;
    int loop= len;
    while(--loop) 
    {
      if( item(pos+loop))
        free(item(pos+loop)); 
    }
    c_Vector::erase(pos,len);
  } 

  void   add     ( lpcstr item)           { c_Vector::add( (ulong)c_strdup(item)); }
  void   insert  ( int pos, lpstr item)   { c_Vector::insert(pos,(ulong)item); }

  int    find    ( lpcstr str) const
  {
    for( int loop=0; loop< m_count; loop++)
    {
      lpstr sz= item(loop);
      if( c_strequ(sz,str)) return loop; 
    }
    return -1;
  }

  int    exists  ( lpcstr str) const      { return (-1==find(str)) ? false:true; }
  lpstr  item    ( int pos)const          { return (lpstr)(c_Vector::item(pos)); }
  lpstr  last    ( int pos=0)const        { return (lpstr)(c_Vector::last(pos)); }
  lpstr          operator[](int pos) const{ return item(pos); }

  c_Strarray&    operator<<(lpcstr item)   { add(item); return *this; }
};

/*  Abstract Array de Void* (Tout inline)
    Attention se sont des POINTEURS
*/
class c_Voidarray: public c_Vector
{
public:
  c_Voidarray ()  { m_size=m_count=0; m_data=NULL; }
  virtual ~c_Voidarray () { erase(); }
  void   erase(int pos=0, int len=0) 
  { 
    if( !m_count) return;
    if( pos<0) pos+= m_count;
    if( pos > m_count) return;
    if( !len) len= m_count;
    if( pos+len > m_count) len= m_count-pos;
    int loop= len;
    while(--loop) 
    {
      if( item(pos+loop))
        delete(item(pos+loop)); 
    }
    c_Vector::erase(pos,len);
  } 

  void   add     ( void* item)            { c_Vector::add( (ulong)item); }
  void   insert  ( int pos, void*item) { c_Vector::insert( pos, (ulong)item); }
  int    find    ( void* item)    { return c_Vector::find((ulong)item); }
  bool   exists  ( void* item)    { return c_Vector::exists((ulong)item); }
  void*  item    ( int pos)    { return (void*)c_Vector::item(pos); }
  void*  last    ( int pos=0)  { return (void*)c_Vector::last(pos); }
  c_Voidarray&   operator<<(void* item)  { add(item); return *this; }
  void*          operator[](int pos)  { return item(pos); }
};


#endif
  


    