/*****************************************************
  file    : c_xem.h
  usage   : classe analyseur de sources XEM 
  date    : 2001
  copyleft: I.R.Maturana <irm@in3activa.com>
  autres  :
  licence : gnu - see www.gnu.org/copyleft/gpl.html
 *****************************************************/


#ifndef _C_XEM_H_
#define _C_XEM_H_

#include  <c_defs.h>
#include  <c_link.h>
#include  <c_file.h>
#include  <c_vector.h>

/*----------------------------------------------------------*/
/*  Constantes
/*----------------------------------------------------------*/

typedef enum{ // Contexte de _eventparse()  et _enventPI
  CX_OK     =0         // 
, CX_ERR_NODATA        // donnes absentes
, CX_ERR_EOF           // Fin de fichier (balise non ferme) 
, CX_ERR_ENDTAG        // balise de fermeture non attendue
, CX_ERR_ESCAPE        // pour <! non suivi de -- ni de [CDATA[ 
, CX_ERR_MEMORY        // memoire insuffisantes
, CX_ERR_FILE          // accs fichier
, CX_ERR_ANY           // ce qu'on veut
}CX_ERROR;

typedef enum{ // Contexte de _eventparse()  et _enventPI
  CX_TYPEUNDEF= 0       // defaut: 
, CX_DECLARATION        // pour declarations de DTD (unused)
, CX_PI_CMD             // processing instruction

, CX_ATTRIBUTE          // attributs de la balise
, CX_ENDBLOCK           // balise de fermeture explicite

, CX_ELEMENT            // balise ouverte
, CX_TAGEND             // balise de fermeture implicite
, CX_COMMENT            // <!-- -->  
, CX_CDATA              // <![DATA[
, CX_TEXT               // node de donnes brutes
} CX_TYPE;

typedef enum{ // Type de sortie
  CX_XEMTYPE= 0       // sortie XEM 
, CX_XMLTYPE          // sortie XML 
} CX_OUTTYPE;           


/*----------------------------------------------------------*/
/*  Node 
/*----------------------------------------------------------*/
#define CAST_XEMNODE(p) ((c_Xemnode*)(p))

class c_Xemnode: public c_Node 
{
friend  class c_Xem;

  // Dfinition XML du node (si standalone, tagname= nom du tag) 
  bool        m_lock;             // Signale un node referm par ENDBLOCK
  CX_TYPE     m_type;             // Type CX_PICMD ou CX_ELEMENT
  c_Str       m_tagname;          // NULL si dclar HEADER, nom du tag si "standalone"
  c_Str       m_cdata;            // Texte jusqu' la balise suivante
  c_Strarray  m_attname;          // noms des attributs 
  c_Strarray  m_attdata;          // valeurs des attributs 

// indirection sur attributs standard
  lpstr       m_anchor;           // indirection sur m_attdata "id='value'"

public: 
  c_Xemnode(): c_Node()	  // constructor
  {
  } 

  /* lock() est utilis pour signaler (=vrai) que le noeud 
    possde une balise de fermeture 
  */
  bool    lock()             { return m_lock; }

  /* Retourne le type du noeud */
  CX_TYPE type   ()          { return m_type   ; }

  /* nom de la balise du noeud */
  lpstr tagname()            { return m_tagname; }

  /* accs aux noms des attributs: 
    pos index de l'attribut  lire
    */
  lpstr attname( size_t pos) { return m_attname[pos]; }

  /* accs aux donnes des attributs: 
    pos index de l'attribut  lire
    */
  lpstr attdata( size_t pos) { return m_attdata[pos]; }

  /* nombre d'attributs du noeud
    */

  int   countatt()           { return m_attdata.count(); }

  /* texte rfrenc par le noeud 
  */
  lpstr cdata()              { return m_cdata;  }

  /*  conversion numrique des donnes du noeud
  */
  long  value()              { return m_cdata.value(); } 

public: 
  /* La classe c_Node est abstraite. c_Node::isvalid() =0 
     La dfinition suivante permet de crer une instance du noeud
    */
  virtual bool isvalid()      { return true; }        

  /* toxem() : sortie du noeud au format XEM
     toxml() : sortie du noeud au format XML
     Les balises de fermeture sont dtectes en explorant les noeuds parent
    */
  bool   toxem(c_Str& buf);      // (re)conversion  XEM 
  bool   toxml(c_Str& buf);      // conversion  XML

  /*  Interprtation du texte contenu dans un noeud
  */
  c_Str& strtoxml(c_Str& buf);   

  /*  Mise en forme du contenu de texte
      Les espaces sont rduits au minimum et les sauts de ligne
      grs d'aprs la prsence d0une minuscule  la ligne suivante
  */
  c_Str& xemstrip(c_Str& buf);   
};


/*----------------------------------------------------------*/
/*  Liste arborescente
/*----------------------------------------------------------*/
class c_Xem : public c_Link
{
  c_File   m_fp;          // interface fichiers

  // Fonctions de parsing
  int   _parsedoc     (); // retourne code error!
  void  _eventdata	  (c_Str& tagname, c_Str& cdata);
  void  _eventatt     (CX_TYPE contexte, c_Str& buf);
  void  _eventdoc     (CX_TYPE contexte, lpstr psz1, lpstr psz2= NULL);

private: 
  /* eventparse() est un remplacement de la fonction c_Link::parse()
     c_Link::parse() traverse toute l'arborescence et appelle eventparse()
     avec pour paramtre, le node courant.
     Voir c_XEM::save() et c_XEM::string() pour l'appel de c_Link::parse()
     */
  virtual bool  eventparse(c_Node* obj); // override, pour save()
  bool          m_isoutfile;             // distingue une sortie fichier ou chane
  c_Str*        m_outstr;                // !NULL si sortie chaine
  c_Xemnode*    m_outnode;               // Reprage pour l'application des tags au parsing

  /* firstElement() permet de retrouver le premier node ELEMENT
     Normalement, les nodes PI sont installs en dbut d'arborescence
     Ce premier node ELEMENT determine la limite d'valuation des balises
     de fermeture.
     */
  c_Xemnode*    firstelement();   // retrouver premier node ELEMENT

public:	  // paramtres de configuration

  CX_OUTTYPE  m_outtype;			  // Type de sortie
  int		  m_withids;			  // 00, 01 si blocs, 11 tous 

  // Gestion des attributs ID. 
  // Noter que m_basids=0=ID0 est rserv au noeud racine
  // vrifi par _parsedoc() 
  int		  m_baseids;			  // compteur de ids aux blocs (si !E)

  c_Str		  m_usertag;			  // balise de bloc obligatoire

public:
  /* Constructeur
	 Defaut CX_XMLTYPE au type de conversion
  */
  c_Xem(): c_Link() 
  { 
	m_outtype = CX_XMLTYPE; 
	m_withids = 0; 
	m_baseids = 1;			  // ID0 est rserv au noeud racine
	errno=CX_OK; 
  };	  

  /* Lecture de fichier XEM/XML. Le type en entre n'est pas ncessaire
     puisque XEM est un sous-ensemble de XML
     */
  bool  load     (lpstr filename);  

  /* Append de fichier XEM/XML. 
	 Comme pour load(), mais en cas de node root, celui-ci est 
	 pralablement limin pour viter l'imbrication de nodes <xem>
     */
  bool  append	 (lpstr filename);  

  /* Sortie fichier de l'arborescence 
     pnode  sous branche de l'arborescence. Null pour sortir toute l'arborescence
	 Note: m_outtype dtermine la syntaxe de sortie (XEM/XML)
     */
  bool  save     (lpstr filename, c_Xemnode* pnode=NULL);  

  /* Sortie chaine de l'arborescence selon le type paramtre. 
     pnode  sous branche de l'arborescence. Null pour sortir toute l'arborescence
	 Note: m_outtype dtermine la syntaxe de sortie (XEM/XML)
     */
  void  string   (c_Str& buffer,  c_Xemnode* pnode=NULL);  


  /* Gestion des messages d'erreur
     errno    code erreur de type enum CX_ERROR
     errstr   message d'erreur en clair
     err()    dfinir une erreur donne	(retourne l'erreur paramtre)
     */
  CX_ERROR errno; 
  c_Str    errstr;         
  int   err( CX_ERROR no, lpstr str) 
        { 
          errno=no; 
          errstr= str; 
          return no;      // pour syntaxe return err(CX_OK, "Ok.");
        }
};


#endif  /* _C_XEM_H_*/
